import tkinter as tk
from tkinter import *
from tkinter import ttk

from ..controller.app_bus import AppBus, AppBusEventType
from ..data.data_manager import DataManager
from ..data.filter import DataFilter

class StatusBar():

    
    def __init__(self, main: Tk, app_bus:AppBus, data_manager:DataManager, row:int):
        self.app_bus = app_bus
        self.data_manager = data_manager
        
        f = Frame(main, bd=1, relief=SUNKEN)
        f.grid(row=row, column=0, columnspan=1, sticky=W+E+N+S)

        self.l_connected = Label(f, borderwidth=1)
        self.l_connected.pack(side=tk.RIGHT, padx=(5, 5), pady=2)
        self.is_connected_handler({'connected':False})

        self.pb = ttk.Progressbar(f, orient=tk.HORIZONTAL, length=160, maximum=100)
        self.pb.pack(side=tk.RIGHT, padx=(5, 0), pady=2)
        self.pb.step(0)

        l = Label(f, text="Device Scan:")
        l.pack(side=tk.RIGHT, padx=(5, 0), pady=2)

        self.l_filter_name = Label(f, text="No Filter")
        self.l_filter_name.pack(side=tk.RIGHT, padx=(0, 0), pady=2)
        self.l = Label(f, text="Active Filter: ")
        self.l.pack(side=tk.RIGHT, padx=(5, 0), pady=2)

        self.l_devices = Label(f, text=self.get_device_count_str())
        self.l_devices.pack(side=tk.LEFT, padx=(5, 0), pady=2)

        self.app_bus.add_event_handler(AppBusEventType.CONNECTION_STATUS_CHANGE, self.is_connected_handler)
        self.app_bus.add_event_handler(AppBusEventType.DEVICE_ITERATION_PROGRESS, self.device_scan_progress_handler)
        self.app_bus.add_event_handler(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, self.update_device_count)
        self.app_bus.add_event_handler(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, self.update_device_count)
        self.app_bus.add_event_handler(AppBusEventType.DEVICE_SCAN_STATUS, self.device_scan_status_handler)
        self.app_bus.add_event_handler(AppBusEventType.SET_DATA_TABLE_FILTER, self.active_filter_name_handler)

        ## initialize
        if self.data_manager.selected_data_filter_name is not None:
            self.active_filter_name_handler(self.data_manager.data_fitlers[self.data_manager.selected_data_filter_name])
            

    def active_filter_name_handler(self, filter:DataFilter):
        if filter is None:
            self.l_filter_name.config(text='No Filter', fg='SystemButtonText')
        elif filter.name is None or filter.name == '':
            self.l_filter_name.config(text='Custom Filter', fg='darkgreen')
        else:
            self.l_filter_name.config(text=filter.name, fg='darkgreen')

    def get_device_count_str(self) -> str:
        count:int = len(self.data_manager.devices)
        fam14s:int = len([d for d in self.data_manager.devices.values() if d.is_fam14()])
        bus_device:int = len([d for d in self.data_manager.devices.values() if d.bus_device])
        decentralized:int = count-bus_device

        return f"Devices: {count}, FAM14s: {fam14s}, Bus Devices: {bus_device}, Decentralized Devices: {decentralized}"

    def update_device_count(self, data):
        self.l_devices.config(text=self.get_device_count_str())

    def device_scan_status_handler(self, status:str):
        if status in ['STARTED', 'FINISHED']:
            self.pb["value"] = 0

    def is_connected_handler(self, data:dict):
        status = data.get('connected')
        if status:
            self.l_connected.config(bg="lightgreen", fg="black", text="Connected")
        else:
            self.l_connected.config(bg="darkred", fg="white", text="Disconnected")

    def device_scan_progress_handler(self, progress:float):
        self.pb["value"] = progress


import xmltodict
import json

from eltakobus.device import *
from eltakobus.util import AddressExpression

from .. import LOGGER
from . import data_helper

from .ha_config_generator import HomeAssistantConfigurationGenerator
from .const import GatewayDeviceType
from .device import Device
from .data_helper import b2s, a2s, a2i, add_addresses, find_device_info_by_device_type

class PCT14DataManager:

    @classmethod
    async def get_devices_from_pct14(cls, filename:str) -> dict:
        devices = {}

        pct14_import = {}
        with open(filename, 'r') as file:
            import_data = xmltodict.parse(file.read())
            pct14_import = import_data['exchange']

        # detect fam14 first
        fam14_device:Device = await cls._create_fam14_device( pct14_import['rootdevice'] )
        devices[fam14_device.external_id] = fam14_device

        for d in pct14_import['devices']['device']:

            dev_size = int(d['header']['addressrange']['#text'])
            for i in range(1, dev_size+1):

                device = await cls._create_device(d, fam14_device, channel=i)
                devices[device.external_id] = device

                for si in device.memory_entries:
                    s:Device = Device.get_decentralized_device_by_sensor_info(si)
                    devices[s.external_id] = s

                    if device.is_ftd14():                    
                        s2:Device = Device.get_decentralized_device_by_sensor_info(si, device.additional_fields['second base id'])
                        devices[s2.external_id] = s2

        return devices


    @classmethod
    async def write_sender_ids_into_existing_pct14_export(cls, source_filename:str, target_filename:str, devices:dict[str,Device], base_id:str):

        with open(source_filename) as xml_file:
            data_dict = xmltodict.parse(xml_file.read())

        fam14_device:Device = await cls._create_fam14_device( data_dict['exchange']['rootdevice'] )
        
        # iterate through PCT14 devices 
        for d in data_dict['exchange']['devices']['device']:

            # interate through device channels
            dev_size = int(d['header']['addressrange']['#text'])
            hw_info = find_device_info_by_device_type(d['name']['#text'])
            if hw_info == {} or 'PCT14-key-function' not in hw_info: 
                print(f"No HW Type fond for: {d['name']['#text']}")
                continue
            function_id = hw_info['PCT14-key-function']

            for i in range(1, dev_size+1):

                # check if device is in eo_man
                external_id = cls._get_external_id(fam14_device, i, d)
                if external_id in devices:
                    _d = devices[external_id]
                    if 'sender' in _d.additional_fields and not cls._is_device_registered(_d, d, i, base_id, function_id):
                        # check if HA sender is registered
                        cls._add_ha_sender_id_into_pct14_xml(base_id, d, d['data']['rangeofid']['entry'], _d)
                    else:
                        LOGGER.debug(f"PCT14 Export Extender: device {_d.name} ('{_d.external_id}' already registered in actuator {d['name']['#text']})")

        # Convert the modified dictionary back to XML
        new_xml = xmltodict.unparse(data_dict, pretty=True)

        # Write the updated XML back to a file
        with open(target_filename, 'w') as xml_file:
            xml_file.write(new_xml)
        LOGGER.debug(f"PCT14 Export Extender: process completed.")

    @classmethod
    def _add_ha_sender_id_into_pct14_xml(cls, base_id, xml_device, xml_entries, device: Device):
        index_range = data_helper.find_device_class_by_name(xml_device['name']['#text']).sensor_address_range
        free_index = -1
        for i in index_range:
            free_index = i
            for n in xml_entries:
                if int(n['entry_number']) == i:
                    free_index = -1
                    break
            if free_index > -1:
                break
            
        sender_id = device.additional_fields['sender']['id']
        if free_index > -1:
            xml_entries.append({
                'entry_number': free_index,
                'entry_id': cls._get_sender_id(base_id, sender_id),
                'entry_function': device.key_function,
                'entry_button': 0,
                'entry_channel': device.channel,
                'entry_value': 0
            })
            LOGGER.debug(f"PCT14 Export Extender: Added sender id '{sender_id}' to PCT14 export for device {device.name} '{device.external_id}'.")
        else:
            LOGGER.warning(f"PCT14 Export Extender:  Cannot add sender id '{sender_id}' entry into device  {device.name} '{device.external_id}'.")

    @classmethod
    def _get_sender_id(cls, base_id: str, sender_id:str):
        sender_offset_id = base_id
        str_id = data_helper.a2s( int(sender_id, base=16) + data_helper.a2i(sender_offset_id) )
        int_id = int(''.join(str_id.split('-')[::-1]), base=16)
        return int_id
            

    @classmethod
    def _is_device_registered(cls, device:Device, pct14_device, channel:int, base_id:str, function_id:int):
        if not isinstance(pct14_device['data']['rangeofid']['entry'], list):
            return False
        
        ha_id = add_addresses(base_id, '00-00-00-'+device.additional_fields['sender']['id'])

        for e in pct14_device['data']['rangeofid']['entry']:

            is_registered = int(e['entry_channel']) & (2**(channel-1)) 
            is_registered = is_registered and (int(e['entry_function']) == function_id)
            is_registered = is_registered and (b2s(cls._convert_sensor_id_to_bytes(e['entry_id'])) == ha_id)
            if is_registered:
                return True
        return False

    @classmethod 
    def _get_external_id(cls, fam14:Device, channel:int, xm_device:dict) -> str:
        id = int(xm_device['header']['address']['#text']) + channel - 1
        return add_addresses(fam14.base_id, a2s(id))

    @classmethod
    async def _create_device(cls, import_obj, fam14:Device, channel:int=1):
        bd = Device()
        bd.additional_fields = {}
        id = int(import_obj['header']['address']['#text']) + channel - 1
        bd.address = a2s( id )
        bd.channel = channel
        bd.dev_size = int(import_obj['header']['addressrange']['#text'])
        bd.use_in_ha = True
        bd.base_id = fam14.base_id
        bd.device_type = import_obj['name']['#text']
        if 'FAM14' in bd.device_type: bd.device_type = GatewayDeviceType.EltakoFAM14.value
        if 'FGW14' in bd.device_type: bd.device_type = GatewayDeviceType.EltakoFGW14USB.value
        if 'FTD14' in bd.device_type: bd.device_type = GatewayDeviceType.EltakoFTD14.value
        int_version = int(import_obj['header']['versionofsoftware']['#text'])
        hex_version = format(int_version, 'X')
        bd.version = f"{hex_version[0]}.{hex_version[1]}"
        bd.key_function = ''
        bd.comment = import_obj['description'].get('#text', '')
        if isinstance(import_obj['channels']['channel'], list):
            for c in import_obj['channels']['channel']:
                if int(c['@channelnumber']) == channel and len(c['@description']) > 0:
                    bd.comment += f" - {c['@description']}"
        else:
            c = import_obj['channels']['channel']
            if int(c['@channelnumber']) == channel and len(c['@description']) > 0:
                    bd.comment += f" - {c['@description']}"

        bd.bus_device = True
        bd.external_id = cls._get_external_id(fam14, channel, import_obj)

        if 'entry' in import_obj['data']['rangeofid']:
            bd.memory_entries = cls._get_sensors_from_xml(bd, import_obj['data']['rangeofid']['entry'])
        else:
            bd.memory_entries = []

        # print(f"{bd.device_type} {bd.address}")
        # print_memory_entires( bd.memory_entries)
        # print("\n")
        bd.name = f"{bd.device_type.upper()} - {bd.address}"
        if bd.comment not in [None, '']:
            bd.name = f"{bd.device_type.upper()} - {bd.comment}"
        else:
            if bd.dev_size > 1:
                bd.name += f" ({bd.channel}/{bd.dev_size})"

        Device.set_suggest_ha_config(bd)

        return bd

    @classmethod
    def _convert_xml_baseid(cls, xml_baseid):
        numbers = []
        for i in range(0,4):
            number = format(int(xml_baseid['baseid_byte_'+str(i)]),'X')
            if len(number) == 1: number = '0'+number
            numbers.append(number)
        return '-'.join(numbers)

        # return f"{}-{format(int(xml_baseid['baseid_byte_1']),'2X')}-{format(int(xml_baseid['baseid_byte_2']),'2X')}-{format(int(xml_baseid['baseid_byte_3']),'2X')}"

    @classmethod
    async def _create_fam14_device(cls, import_obj):
        bd = Device()
        bd.additional_fields = {}
        xml_baseid = import_obj['rootdevicedata']['baseid']
        bd.address = cls._convert_xml_baseid(xml_baseid)
        bd.channel = 1
        bd.dev_size = int(import_obj['header']['addressrange']['#text'])
        bd.use_in_ha = True
        bd.base_id = bd.address
        bd.device_type = FAM14
        bd.device_type = GatewayDeviceType.EltakoFAM14.value

        bd.version = "" #TODO: '.'.join(map(str,device.version))
        bd.key_function = ''
        bd.comment = import_obj['description']['#text']
        bd.bus_device = True
        bd.external_id = bd.base_id

        bd.memory_entries = []
        if bd.comment not in [None, '']:
            bd.name = f"{bd.device_type} - {bd.comment}"
        else:
            bd.name = f"{bd.device_type} - {bd.address}"
        
        Device.set_suggest_ha_config(bd)

        return bd
    

    @classmethod
    def _convert_sensor_id_to_bytes(cls, id:str):
        hex_rep = format(int(id), 'X')
        if len(hex_rep) % 2 == 1:
            hex_rep = "0"+hex_rep
        hex_rep = [hex_rep[i:i+2] for i in range(0, len(hex_rep), 2)]
        hex_rep = hex_rep[::-1]
        hex_string = '-'.join(hex_rep)
        return AddressExpression.parse(hex_string)[0]


    @classmethod
    def _get_sensors_from_xml(cls, device:Device, sensors) -> list[SensorInfo]:
        return [cls._get_sensor_from_xml(device, s) for s in sensors]

    @classmethod
    def _get_sensor_from_xml(cls, device:Device, sensor) -> list[SensorInfo]:

        hw_info = find_device_info_by_device_type(device.device_type)
        if hw_info:
            hw_info = hw_info['hw-type']
        else:
            hw_info = None

        return SensorInfo(
                dev_type = hw_info,
                sensor_id = cls._convert_sensor_id_to_bytes(sensor['entry_id']),
                dev_adr = a2i(device.address).to_bytes(4, byteorder = 'big'),
                key = int(sensor['entry_button']),
                dev_id = device.address,
                key_func = int(sensor['entry_function']),
                channel = int(sensor['entry_channel']),
                in_func_group=None,
                memory_line = int(sensor['entry_number'])
                )


from .device import Device

class DataFilter():

    def __init__(self, name:str, 
                 global_filter:list[str]=[], 
                 device_address_filter:list[str]=[],
                 device_external_address_filter:list[str]=[],
                 device_type_filter:list[str]=[],
                 device_eep_filter:list[str]=[]):
        self.name = name
        self.global_filter = global_filter
        self.device_address_filter = device_address_filter
        self.device_external_address_filter = device_external_address_filter
        self.device_type_filter = device_type_filter
        self.device_eep_filter = device_eep_filter

    
    def filter_device(self, device:Device):
        # check address
        for f in self.device_address_filter + self.global_filter:
            if device.address and f.upper() in device.address.upper():
                return True
        
        # check external id
        for f in self.device_external_address_filter + self.global_filter:
            if device.external_id and f.upper() in device.external_id.upper():
                return True
            
        # check device type
        for f in self.device_type_filter + self.global_filter:
            if device.device_type and f.upper() in device.device_type.upper():
                return True
        
        # check eep
        for f in self.device_eep_filter + self.global_filter:
            if device.eep and f.upper() in device.eep.upper():
                return True
            
        for f in self.global_filter:
            # key function
            if device.key_function and f.upper() in device.key_function.upper():
                return True
            
            # comment
            if device.comment and f.upper() in device.comment.upper():
                return True
            
            # version
            if device.version and f.upper() in device.version.upper():
                return True
            
            # ha platform
            if device.ha_platform and f.upper() in device.ha_platform.upper():
                return True

            if self.find_in_dict(device.additional_fields, f.upper()):
                return True

        return False
    
    
    def find_in_dict(self, additional_fields:dict, filter:str) -> bool:
        for key, value in additional_fields.items():
            if isinstance(value, dict):
                if self.find_in_dict(value, filter):
                    return True
            elif filter in str(value).upper():
                return True
        return False


import unittest

from eo_man import load_dep_homeassistant
load_dep_homeassistant()

from eo_man.data.pct14_data_manager import PCT14DataManager

class TestPCT14Import(unittest.IsolatedAsyncioTestCase):

    async def test_import(self):
        devices = await PCT14DataManager.get_devices_from_pct14('./tests/resources/20240925_PCT14_export_test.xml') 

        self.assertEqual(len(devices), 64)
        self.assertEqual(len([d for d in devices.values() if d.bus_device]), 15)


    async def test_write_sender_ids_into_existing_pct_export(self):
        devices = await PCT14DataManager.get_devices_from_pct14('./tests/resources/20240925_PCT14_export_test.xml') 

        await PCT14DataManager.write_sender_ids_into_existing_pct14_export(
            source_filename='./tests/resources/20240925_PCT14_export_test.xml',
            target_filename='./tests/resources/20240925_PCT14_export_test_GENERATED.xml',
            devices=devices,
            base_id='00-00-B0-00')


import os
import asyncio
import threading

from tkinter import *
from tkinter import ttk
from tkinter import filedialog
import logging
from tkinter import messagebox
import webbrowser

from ..controller.app_bus import AppBus, AppBusEventType
from ..controller.serial_controller import SerialController

from ..data.device import Device
from ..data.data_manager import DataManager
from ..data.ha_config_generator import HomeAssistantConfigurationGenerator
from ..data.pct14_data_manager import PCT14DataManager

from ..icons.image_gallary import ImageGallery

from .eep_checker_window import EepCheckerWindow
from .about_window import AboutWindow
from .device_info_window import DeviceInfoWindow
from .send_message_window import SendMessageWindow
from . import DEFAULT_WINDOW_TITLE

class MenuPresenter():

    def __init__(self, main: Tk, app_bus: AppBus, data_manager: DataManager, serial_controller:SerialController):
        self.main = main
        self.app_bus = app_bus
        self.data_manager = data_manager
        self.serial_controller = serial_controller
        self.ha_conf_gen = HomeAssistantConfigurationGenerator(app_bus, data_manager)
        self.remember_latest_filename = ""

        self.send_message_window = None

        menu_bar = Menu(main)
        file_menu = Menu(menu_bar, tearoff=False)
        
        # filemenu.add_command(label="New")
        load_file_icon = ImageGallery.get_open_folder_icon(size=(16,16))
        file_menu.add_command(label="Load File...", 
                              image=load_file_icon,
                              compound=LEFT,
                              command=self.load_file, 
                              accelerator="Ctrl+O")
        file_menu.add_command(label="Import From File...", 
                              image=load_file_icon,
                              compound=LEFT,
                              command=self.import_from_file, 
                              accelerator="Ctrl+I")
        file_menu.add_separator()
        file_menu.add_command(label="Load PCT14 Export...", 
                              image=ImageGallery.get_pct14_icon(size=(16,16)),
                              compound=LEFT,
                              command=self.load_pct14_export)
        file_menu.add_command(label="Extend existing PCT14 Export...", 
                              image=ImageGallery.get_pct14_icon(size=(16,16)),
                              compound=LEFT,
                              command=self.extend_pct14_export)
        file_menu.add_separator()
        file_menu.add_command(label="Remove all Device from Table", 
                              image=ImageGallery.get_clear_icon(size=(16,16)),
                              compound=LEFT,
                              command=self.remove_all_device_from_table)
        file_menu.add_separator()
        ha_icon = ImageGallery.get_ha_logo(size=(16,16))
        file_menu.add_command(label="Export Home Assistant Configuration", 
                              image=ha_icon, 
                              compound=LEFT, 
                              command=self.export_ha_config, 
                              accelerator="Ctrl+E")
        file_menu.add_command(label="Export Home Assistant Configuration as ...", 
                              image=ha_icon, 
                              compound=LEFT, 
                              command=lambda: self.export_ha_config(save_as=True), 
                              accelerator="Ctrl+SHIFT+E")
        file_menu.add_separator()
        save_icon = ImageGallery.get_save_file_icon((16,16))
        file_menu.add_command(label="Save", 
                              image=save_icon,
                              compound=LEFT,
                              command=self.save_file, 
                              accelerator="Ctrl+S")
        save_as_icon = ImageGallery.get_save_file_as_icon((16,16))
        file_menu.add_command(label="Save as...", 
                             image=save_as_icon,
                             compound=LEFT,
                             command=lambda: self.save_file(save_as=True), 
                             accelerator="Ctrl+SHIFT+S")
        file_menu.add_separator()
        file_menu.add_command(label="Exit", command=main.quit, accelerator="ALT+F4")

        menu_bar.add_cascade(menu=file_menu, 
                                  label="File",
                                  accelerator="ALT+F")
        

        ha_menu = Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label="Home Assistant", menu=ha_menu)
        ha_menu.add_command(label="Reset to suggested HA properties.",
                            command=self.reset_to_suggested_ha_properties)


        tool_menu = Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label="Tools", 
                             menu=tool_menu)
        tool_menu.add_command(label="EEP Checker",
                              command=lambda: EepCheckerWindow(main))
        tool_menu.add_command(label="Send Message", 
                              image=ImageGallery.get_forward_mail((16,16)),
                              compound=LEFT,
                              command=self.show_send_message_window)
        tool_menu.add_command(label="Message Log Analyser", 
                              command=lambda: messagebox.showinfo("Message Log Analyser", "Will be available soon!"))

        help_menu = Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label="Help", menu=help_menu)
        
        github_icon = ImageGallery.get_github_icon((16,16))
        docs_menu = Menu(help_menu, tearoff=False)
        help_menu.add_command(label="Supported Devices",
                             compound=LEFT,
                             command=lambda: DeviceInfoWindow(main) )
        docs_menu.add_separator()
        help_menu.add_cascade(menu=docs_menu, 
                             label="Documentation")
        docs_menu.add_command(label="EnOcean Device Manager ...", 
                             image=github_icon,
                             compound=LEFT,
                             command=self.open_eo_man_documentation)
        docs_menu.add_separator()
        docs_menu.add_command(label="Home Assistant Eltako Integration ...", 
                             image=github_icon,
                             compound=LEFT,
                             command=lambda: webbrowser.open_new(r"https://github.com/grimmpp/home-assistant-eltako/tree/main/docs"))

        repos_menu = Menu(help_menu, tearoff=0)
        help_menu.add_cascade(menu=repos_menu, 
                             label="GitHub Repositories")
        repos_menu.add_command(label="EnOcean Device Manager ...", 
                              image=github_icon,
                              compound=LEFT,
                              command=lambda: webbrowser.open_new(r"https://github.com/grimmpp/enocean-device-manager"))
        repos_menu.add_separator()
        repos_menu.add_command(label="Home Assistant Eltako Integration ...", 
                               image=github_icon,
                               compound=LEFT,
                               command=self.open_eo_man_repo)
        repos_menu.add_separator()
        repos_menu.add_command(label="Eltako14Bus Communication Library ...", 
                               image=github_icon,
                               compound=LEFT,
                               command=lambda: webbrowser.open_new(r"https://github.com/grimmpp/eltako14bus"))

        about_icon = ImageGallery.get_about_icon((16,16))
        help_menu.add_command(label="About...", 
                              image=about_icon,
                              compound=LEFT,
                              command=lambda: AboutWindow(main), 
                              accelerator="F1")
        

        main.config(menu=menu_bar)

        # add shortcuts globally
        main.bind('<Control-o>', lambda e: self.load_file())
        main.bind('<Control-i>', lambda e: self.import_from_file())
        main.bind('<Control-e>', lambda e: self.export_ha_config())
        main.bind('<Control-Shift-E>', lambda e: self.export_ha_config(save_as=True))
        main.bind('<Control-s>', lambda e: self.save_file())
        main.bind('<Control-Shift-S>', lambda e: self.save_file(save_as=True))
        main.bind('<F1>', lambda e: AboutWindow(main))
        


    def save_file(self, save_as:bool=False):
        try:
            if save_as or not os.path.isfile(self.remember_latest_filename):
            # initial_dir = os.path.dirname(self.remember_latest_filename)

                filename = filedialog.asksaveasfilename(initialdir=self.remember_latest_filename, 
                                                    title="Save Application State",
                                                    filetypes=[("EnOcean Device Manager", "*.eodm"), ("EnOcean Device Manager", "*.yaml")],
                                                    defaultextension=".eodm")
                if not filename:
                    return
                
                if not filename.endswith('.eodm') and not filename.endswith('.yaml'):
                    filename += '.eodm'
                self.remember_latest_filename = filename

            self.data_manager.write_application_data_to_file(self.remember_latest_filename)
            # with open(self.remember_latest_filename, 'wb') as file:
            #     pickle.dump( self.data_manager.devices, file)
            
            self.main.title(f"{DEFAULT_WINDOW_TITLE} ({os.path.basename(self.remember_latest_filename)})")
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Save to File: '{self.remember_latest_filename}'", 'color': 'red'})

        except Exception as e:
            msg = f"Saving application configuration to file '{filename}' failed!"
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
            logging.exception(msg, exc_info=True)


    def import_from_file(self):
        filename = None
        
        if not self.remember_latest_filename:
            self.remember_latest_filename = os.path.expanduser('~')

        initial_dir = os.path.dirname(self.remember_latest_filename)
        filename = filedialog.askopenfilename(initialdir=initial_dir, 
                                                title="Load Application State",
                                                filetypes=[("EnOcean Device Manager", "*.eodm"), ("EnOcean Device Manager", "*.yaml")],
                                                defaultextension=".eodm") #, ("configuration", "*.yaml"), ("all files", "*.*")))

        if not filename:
            return None

        def load():
            try:
                self.data_manager.load_application_data_from_file(filename)
                # with open(filename, 'rb') as file:
                #     self.data_manager.load_devices( pickle.load(file) )
            except Exception as e:
                msg = f"Loading application configuration file '{filename}' failed!"
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
                logging.exception(msg, exc_info=True)

        t = threading.Thread(target=load)
        t.start()

        return filename
        
        
    def load_file(self, filename:str=None):
        if filename == None:
            filename = self.import_from_file()

        if filename:
            self.app_bus.fire_event(AppBusEventType.LOAD_FILE, {})
            self.remember_latest_filename = filename

            self.main.title(f"{DEFAULT_WINDOW_TITLE} ({os.path.basename(self.remember_latest_filename)})")
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Load File: '{self.remember_latest_filename}'", 'color': 'red'})

    
    remember_latest_ha_config_filename:str=None
    def export_ha_config(self, save_as:bool=False):
        try:
            if save_as or not self.remember_latest_ha_config_filename:
            # initial_dir = os.path.dirname(self.remember_latest_filename)

                filename = filedialog.asksaveasfilename(initialdir=self.remember_latest_filename, 
                                                    title="Save Home Assistant Configuration",
                                                    filetypes=[("Home Assistant Configuration", "*.yaml")],
                                                    defaultextension=".yaml")
                if not filename:
                    return

                if not filename.endswith('.yaml'):
                    filename += '.yaml'
                self.remember_latest_ha_config_filename = filename

            


            msg = f"Export Home Assistant configuration into {self.remember_latest_ha_config_filename}"
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'color': 'red', 'log-level': 'INFO'})
            

            error = self.ha_conf_gen.perform_tests()
            if error is not None:
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': error, 'log-level': 'ERROR', 'color': 'red'})
                if not messagebox.askyesno(title="Error in Checking Configuration Data!", message=error+"\n\nDo you want to try to continue?"):
                    self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': 'Exporting HA Configuration cancelled by user.', 'log-level': 'INFO'})
                    return

            self.ha_conf_gen.save_as_yaml_to_file(self.remember_latest_ha_config_filename)

            msg = f"Export finished!"
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'color': 'grey', 'log-level': 'DEBUG'})
            
            msg = f"Home Assistant Configuration was successfully generated into file: \n\n{self.remember_latest_filename}"
            messagebox.showinfo("Successful Export", msg)

        except Exception as e:
            msg = 'Exporting Home Assistant Configuration failed!'
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
            logging.exception(msg, exc_info=True)

            messagebox.showwarning(title=msg, message=str(e) )


    def reset_to_suggested_ha_properties(self):
        yes = messagebox.askyesno("Reset to suggested HA properties.", "Do you want to reset Home Assistant relevant properties of all devices to initial values?")
        if yes:
            for d in self.data_manager.devices.values():
                Device.set_suggest_ha_config(d)
                if d.is_bus_device():
                    self.app_bus.fire_event(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, d)
                
                    for me in d.memory_entries:
                        s = Device.get_decentralized_device_by_sensor_info(me, d.base_id)
                        if s.external_id in self.data_manager.devices:
                            Device.merge_devices(self.data_manager.devices[s.external_id], s)
                            self.app_bus.fire_event(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, s)


    def open_eo_man_repo(self):
        webbrowser.open_new(r"https://github.com/grimmpp/enocean-device-manager")
        
    def open_eo_man_documentation(self):
        webbrowser.open_new(r"https://github.com/grimmpp/enocean-device-manager/tree/main/docs")

    def show_send_message_window(self):
        if self.send_message_window is None:
            self.send_message_window = SendMessageWindow(self.main, self.app_bus, self.data_manager, self.serial_controller)
        self.send_message_window.show_window()

    def load_pct14_export(self):
        filename = filedialog.askopenfilename(initialdir=os.path.expanduser('~'), 
                                              title="Load PCT14 Export",
                                                filetypes=[("PCT14 Export", "*.xml")],
                                                defaultextension=".xml")
        
        if not filename:
            return None

        def load():
            try:
                devices = asyncio.run( PCT14DataManager.get_devices_from_pct14(filename))
                self.data_manager.load_devices(devices)

            except Exception as e:
                msg = f"Loading PCT14 Export '{filename}' failed!"
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
                logging.exception(msg, exc_info=True)

        t = threading.Thread(target=load)
        t.start()

        return filename
    
    def extend_pct14_export(self):
        source_filename = filedialog.askopenfilename(initialdir=os.path.expanduser('~'), 
                                              title="Extend existing PCT14 Export",
                                              filetypes=[("PCT14 Export", "*.xml")],
                                              defaultextension=".xml")        
        if not source_filename:
            return None

        def load():
            try:
                target_filename = source_filename.replace('.xml', '_extended.xml')
                asyncio.run( PCT14DataManager.write_sender_ids_into_existing_pct14_export(
                    source_filename, target_filename, 
                    self.data_manager.devices,
                    HomeAssistantConfigurationGenerator.LOCAL_SENDER_OFFSET_ID))

                msg = f"Extended PCT14 Export '{target_filename}'!"
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'INFO', 'color': 'green'})

            except Exception as e:
                msg = f"Extend PCT14 Export '{target_filename}' failed!"
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
                logging.exception(msg, exc_info=True)

        t = threading.Thread(target=load)
        t.start()

    def remove_all_device_from_table(self):
        if self.serial_controller.is_serial_connection_active():
            self.serial_controller.stop_serial_connection()

        self.data_manager.devices = {}
        self.data_manager.load_devices({})
        self.app_bus.fire_event(AppBusEventType.SET_DATA_TABLE_FILTER, None)
        self.app_bus.fire_event(AppBusEventType.SELECTED_DEVICE, None)



from typing import Dict, List

from zeroconf import Zeroconf, ServiceBrowser, ServiceInfo

import socket

from .app_bus import AppBus, AppBusEventType
from ..data.const import GatewayDeviceType, get_gateway_type_by_name
from ..data import data_helper 

class LanServiceDetector:

    def __init__(self, app_bus: AppBus, update_callback) -> None:
        
        self.app_bus = app_bus
        self._update_callback = update_callback
        self.service_reg_lan_gw = {}
        self.service_reg_virt_lan_gw = {}
        self._start_service_discovery()


    def find_mdns_service_by_ip_address(self, address:str):
        service_list:List[Dict] = {}
        service_list.update(self.service_reg_lan_gw)
        service_list.update(self.service_reg_virt_lan_gw)

        for s in service_list.values():
            dns_name = f"{s['hostname'][:-1]}:{s['port']}"
            ip_address = f"{s['address']}:{s['port']}"
            if dns_name == address or ip_address == address:
                return s['name']
                
        return None

    def __del__(self):
        self.zeroconf.close()    

    def add_service(self, zeroconf: Zeroconf, type, name):
        try:
            info:ServiceInfo = zeroconf.get_service_info(type, name)
            obj = {'name': name, 'type': type, 'address': socket.inet_ntoa(info.addresses[0]), 'port': info.port, 'hostname': info.server}
            msg = f"Detected Network Service: {name}, type: {type}, address: {obj['address']}, port: {obj['address']}, hostname: {obj['address']}"
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'INFO', 'color': 'grey'})
            
            for mdns_name in data_helper.MDNS_SERVICE_2_GW_TYPE_MAPPING:
                if mdns_name in name:
                    gw_type: GatewayDeviceType = data_helper.MDNS_SERVICE_2_GW_TYPE_MAPPING[mdns_name]
                    if gw_type == GatewayDeviceType.LAN:
                        self.service_reg_lan_gw[name] = obj
                        break
                    elif gw_type == GatewayDeviceType.LAN_ESP2:
                        self.service_reg_virt_lan_gw[name] = obj
                        break

        except:
            pass

        self._update_callback()

    def remove_service(self, zeroconf, type, name):
        if name in self.service_reg_lan_gw:
            del self.service_reg_lan_gw[name]
        if name in self.service_reg_virt_lan_gw:
            del self.service_reg_virt_lan_gw[name]

        self._update_callback()

    def update_service(self, zeroconf, type, name):
        self._update_callback()

    def _start_service_discovery(self):
        self.zeroconf = Zeroconf()
        for mdns_type in data_helper.KNOWN_MDNS_SERVICES.values():
            ServiceBrowser(self.zeroconf, mdns_type, self)

    def get_virtual_network_gateway_service_endpoints(self):
        return [f"{s['hostname'][:-1]}:{s['port']}"  for s in self.service_reg_virt_lan_gw.values()]
    
    def get_lan_gateway_endpoints(self):
        return [f"{s['address']}:{s['port']}"  for s in self.service_reg_lan_gw.values()]


import threading
import tkinter as tk
from tkinter import ttk
import os
from tkinter import *
from tkinter import messagebox
from PIL import Image, ImageTk
from idlelib.tooltip import Hovertip
import webbrowser
import subprocess

from eo_man import LOGGER

from .donation_button import DonationButton
from .menu_presenter import MenuPresenter
from ..icons.image_gallary import ImageGallery
from ..data.app_info import ApplicationInfo as AppInfo


class ToolBar():
    # icon list
    # https://commons.wikimedia.org/wiki/Comparison_of_icon_sets
    def __init__(self, main: Tk, menu_presenter:MenuPresenter, row:int):
        self.main = main

        f = Frame(main, bd=1)#, relief=SUNKEN)
        f.grid(row=row, column=0, columnspan=1, sticky=W+E+N+S, pady=2, padx=2)

        b = self._create_img_button(f, "Save to current file", ImageGallery.get_save_file_icon(), menu_presenter.save_file )
        b = self._create_img_button(f, "Save as file", ImageGallery.get_save_file_as_icon(), lambda: menu_presenter.save_file(save_as=True) )
        b = self._create_img_button(f, "Open file", ImageGallery.get_open_folder_icon(), menu_presenter.load_file)
        
        self._add_separater(f)
        
        b = self._create_img_button(f, "Export Home Assistant Configuration", ImageGallery.get_ha_logo(), menu_presenter.export_ha_config)

        self._add_separater(f)

        b = self._create_img_button(f, "Send Message", ImageGallery.get_forward_mail(), menu_presenter.show_send_message_window)

        # placed at the right end
        b = DonationButton(f, relief=GROOVE, small_icon=True).pack(side=RIGHT, padx=(0,2), pady=2)
        b = self._create_img_button(f, "GitHub: EnOcean Device Manager Repository", ImageGallery.get_github_icon(), menu_presenter.open_eo_man_repo)
        b.pack(side=RIGHT, padx=(0,2), pady=2)
        b = self._create_img_button(f, "GitHub: EnOcean Device Manager Documentation", ImageGallery.get_help_icon(), menu_presenter.open_eo_man_documentation)
        b.pack(side=RIGHT, padx=(0,2), pady=2)

        if not AppInfo.is_version_up_to_date():
            new_v = AppInfo.get_lastest_available_version()
            b = self._create_img_button(f, f"New Software Version 'v{new_v}' is available.", ImageGallery.get_software_update_available_icon(), self.show_how_to_update)
            b.pack(side=RIGHT, padx=(0,2), pady=2)

    def _add_separater(self, f:Frame):
        ttk.Separator(f, orient=VERTICAL).pack(side=LEFT, padx=4)


    def _create_img_button(self, f:Frame, tooltip:str, image:ImageTk.PhotoImage, command) -> Button:
        b = ttk.Button(f, image=image, cursor="hand2", command=command) #, relief=GROOVE)
        Hovertip(b,tooltip,300)
        b.image = image
        b.pack(side=LEFT, padx=(2,0), pady=2)
        return b
    
    def show_how_to_update(self):
        base_path = os.environ.get('VIRTUAL_ENV', '')
        if base_path != '':
            base_path = os.path.join(base_path, 'Scripts')
        
        new_version = AppInfo.get_lastest_available_version()
        
        msg  = f"A new version 'v{new_version}' of 'EnOcean Device Manager' is available. \n\n"
        msg += f"You can update this application by entering \n"
        msg += f"'{os.path.join(base_path, 'pip')}' install eo_man --upgrade'\n"
        msg += f"into the command line."

        messagebox.showinfo("Update Available", msg)


import unittest
import sys 
import os
import asyncio

file_dir = os.path.join( os.path.dirname(__file__), '..', 'eo_man', 'data')
sys.path.append(file_dir)
__import__('homeassistant') 


from tests.mocks import AppBusMock

from eo_man.controller.serial_port_detector import SerialPortDetector



class TestDetectingUsbDevices(unittest.TestCase):

    @unittest.skip("Not yet compatible with Linux")
    def test_print_device_info(self):
        SerialPortDetector.print_device_info()

    @unittest.skip("Not yet compatible with Linux")
    def test_port_detection(self):
        spd = SerialPortDetector(AppBusMock())
        mapping = asyncio.run( spd.async_get_gateway2serial_port_mapping() )
        pass


import time
import unittest
import os
import pickle

from eo_man import load_dep_homeassistant
load_dep_homeassistant()


from eo_man.controller.app_bus import AppBus

from eo_man.data.data_manager import DataManager
from eo_man.data.application_data import ApplicationData


class TestLoadAndStoreAppConfig(unittest.TestCase):

    def test_load(self):
        app_bus = AppBus()
        dm = DataManager(app_bus)

        filename = os.path.join( os.path.dirname(__file__), 'resources', 'test_app_config_1.eodm')
        self.assertTrue(os.path.isfile(filename), f"{filename} is no valid filename.")
        dm.load_application_data_from_file(filename)

        self.assertEqual(22, len(dm.devices), "Loaded device count does not match!")
        

    def test_load_save_load(self):
        app_bus = AppBus()
        dm = DataManager(app_bus)

        filename = os.path.join( os.path.dirname(__file__), 'resources', 'test_app_config_1.eodm')
        self.assertTrue(os.path.isfile(filename), f"{filename} is no valid filename.")
        app_data:ApplicationData = dm.load_application_data_from_file(filename)

        filename2 = os.path.join( os.path.dirname(__file__), 'resources', 'test_app_config_1_temp.eodm')
        dm.write_application_data_to_file(filename2)

        self.assertTrue(os.path.isfile(filename2), f"{filename2} is no valid filename.")
        app_data2:ApplicationData = dm.load_application_data_from_file(filename2)
        
        filename3 = os.path.join( os.path.dirname(__file__), 'resources', 'test_app_config_1_temp2.eodm')
        dm.write_application_data_to_file(filename3)

        self.assertTrue(os.path.isfile(filename3), f"{filename3} is no valid filename.")
        app_data3:ApplicationData = dm.load_application_data_from_file(filename3)

        # compare app_data
        self.assertEqual(len(app_data.devices),                     len(app_data2.devices))
        self.assertEqual(len(app_data.data_filters),                len(app_data2.data_filters))
        self.assertEqual(len(app_data.selected_data_filter_name),   len(app_data2.selected_data_filter_name))
        # self.assertEqual(len(app_data.application_version),       len(app_data2.application_version))   # will not match after version change

        self.assertEqual(len(app_data2.devices),                    len(app_data3.devices))
        self.assertEqual(len(app_data2.data_filters),               len(app_data3.data_filters))
        self.assertEqual(len(app_data2.selected_data_filter_name),  len(app_data3.selected_data_filter_name))
        self.assertEqual(len(app_data2.application_version),        len(app_data3.application_version))
        
        # "binary" check
        self.assertEqual(pickle.dumps(app_data2), pickle.dumps(app_data3))
        


from tkinter import *
import tkinter as tk
from tkinter import Tk, ttk

from eltakobus.message import *
from eltakobus.eep import A5_08_01

from ..data import data_helper


class EepCheckerWindow():

    def __init__(self, main:Tk):
        popup = Toplevel(main, padx=4, pady=4)
        popup.wm_title("EEP Checker")
        self.popup = popup

        row = 0
        self.l_telegram = ttk.Label(popup, text="A5 5A")
        self.l_telegram.grid(row=row, column=0, sticky=EW, columnspan=2)

        row += 1
        l = ttk.Label(popup, text="EEP: ")
        l.grid(row=row, column=0, sticky=W)

        self.cb_eep = ttk.Combobox(popup, state="readonly", width="14") 
        self.cb_eep['values'] = data_helper.get_all_eep_names()
        self.cb_eep.set(A5_08_01.eep_string)
        self.cb_eep.grid(row=row, column=1, sticky=W)
        self.cb_eep.bind('<<ComboboxSelected>>', lambda e: [self.set_message_type(), self.show_eep_values(e)])


        row += 1
        l = ttk.Label(popup, text="Message Type: ")
        l.grid(row=row, column=0, sticky=W)

        self.l_msg_type = ttk.Label(popup, text='')
        self.l_msg_type.grid(row=row, column=1, sticky=W)
        # self.cb_msg_type = ttk.Combobox(popup, state="readonly", width="14") 
        # self.cb_msg_type['values'] = ['RPS (Org = 0x05)', '1BS (Org = 0x06)', '4BS (Org = 0x7)']
        # self.cb_msg_type.set('4BS (Org = 0x7)')
        # self.cb_msg_type.grid(row=row, column=1, sticky=W)
        # self.cb_msg_type.bind('<<ComboboxSelected>>', self.show_message)


        row += 1
        l = ttk.Label(popup, text="Data: ")
        l.grid(row=row, column=0, sticky=W)

        self.e_data = ttk.Entry(popup, width="12") 
        self.e_data.insert(END, "aa 80 76 0f")
        self.e_data.bind_all('<Key>', self.show_eep_values)
        self.e_data.grid(row=row, column=1, sticky=W)


        row += 1
        l = ttk.Label(popup, text="EEP Values: ")
        l.grid(row=row, column=0, sticky=W, columnspan=2)


        row += 1
        self.l_result = ttk.Label(popup, text="", font=("Arial", 11))
        self.l_result.grid(row=row, column=0, sticky=W, columnspan=2)


        row += 1
        l = ttk.Label(popup, text="")
        l.grid(row=row, column=0, sticky=W, columnspan=2)


        row += 1
        b = ttk.Button(popup, text="Close", command=popup.destroy)
        b.bind('<Return>', lambda e: popup.destroy())
        b.grid(row=row, column=0, columnspan=2, sticky=E)

        self.show_eep_values(None)

        popup.wm_attributes('-toolwindow', 'True')

        def center():
            w = popup.winfo_width()
            h = popup.winfo_height()
            x = main.winfo_x() + main.winfo_width()/2 - w/2
            y = main.winfo_y() + main.winfo_height()/2 - h/2
            popup.geometry('%dx%d+%d+%d' % (w, h, x, y))

        popup.after(10, center)

        main.wait_window(popup)


    
    def set_message_type(self):
        try:
            msg = Regular4BSMessage(b'\x00\x00\x00\x00', 0x00, b'\x00\x00\x00\x00', False)
            eep = data_helper.find_eep_by_name(self.cb_eep.get())
            eep_values = data_helper.get_values_for_eep(eep, msg)
            self.l_msg_type.config(text= '4BS (Org = 0x7)')
            return '4BS (Org = 0x7)'
        except:
            pass

        try:
            msg = Regular1BSMessage(b'\x00\x00\x00\x00', 0x00, b'\x00', False)
            eep = data_helper.find_eep_by_name(self.cb_eep.get())
            eep_values = data_helper.get_values_for_eep(eep, msg)
            self.l_msg_type.config(text= '1BS (Org = 0x06)')
            return '1BS (Org = 0x06)'
        except:
            pass

        try:
            msg = RPSMessage(b'\x00\x00\x00\x00', 0x00, b'\x00', False)
            eep = data_helper.find_eep_by_name(self.cb_eep.get())
            eep_values = data_helper.get_values_for_eep(eep, msg)
            self.l_msg_type.config(text= 'RPS (Org = 0x05)')
            return 'RPS (Org = 0x05)'
        except:
            pass


    def get_data_from_entry(self):
        text = self.e_data.get().upper()
        hex_string = '0x'
        for c in text:
            if c in ['0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F']:
                hex_string += c
        return int(hex_string, 16).to_bytes(4, 'big') 


    def show_eep_values(self, event):
        msg_type = self.set_message_type()
        try:
            data = self.get_data_from_entry()
            
            if '4BS' in msg_type:
                msg = Regular4BSMessage(b'\x00\x00\x00\x00', 0x00, data, False)
            elif '1BS' in msg_type:
                data = data[0:1]
                msg = Regular1BSMessage(b'\x00\x00\x00\x00', 0x00, data, False)
            elif 'RPS' in msg_type:
                msg = RPSMessage(b'\x00\x00\x00\x00', 0x00, data, False)
                data = data[0:1]

            msg_text = b2a(msg.serialize()).upper()
            self.l_telegram.config(text=msg_text)
        except:
            self.l_telegram.config(text='')

        try:
            eep = data_helper.find_eep_by_name(self.cb_eep.get())
            eep_values = data_helper.get_values_for_eep(eep, msg)
            self.l_result.config(text='\n'.join(eep_values).replace('_',' '))

        except:
            self.l_result.config(text='')


import unittest

from eltakobus.message import ESP2Message, Regular4BSMessage

class TestBaseIdUtils(unittest.TestCase):
    
    def test_add_addresses(self):
        a1 = b'\xff\x00\xff\x00'
        a2 = b'\x00\xff\x00\xff'
        
        r = self.add_two_addresses(a1, a2)
        self.assertEqual(r, b'\xff\xff\xff\xff')
        
        a1 = b'\xff\x00\x00\xff'
        r = self.add_two_addresses(a1, a2)
        self.assertEqual(r, b'\xff\xff\x01\xfe')
        
    def add_two_addresses(self, a1, a2):
        # return bytes((a + b) & 0xFF for a, b in zip(a1, a2))
        return (int.from_bytes(a1, 'big') + int.from_bytes(a2, 'big')).to_bytes(4, byteorder='big')
    
    
    def test_add_base_id_to_local_message(self):
        base_id = b'\xff\x0C\x0b\x80'
        msg = Regular4BSMessage(b'\x00\x00\x00\x05', 50, b'\x08\x08\x08\x08', True)
        address = self.add_two_addresses(base_id, msg.address)
        msg2 = ESP2Message( msg.body[:8] + address + msg.body[12:] )
        
        self.assertTrue(base_id not in msg.serialize())
        self.assertTrue(base_id not in msg2.serialize())
        
        self.assertTrue(address not in msg.serialize())
        self.assertTrue(address in msg2.serialize())
        
        
    def test_address_checks(self):
        address = b'\x00\x00\x00\x08'
        self.assertEqual( address[0], 0)
        self.assertEqual( address[0:2], b'\x00\x00')
        self.assertEqual( int.from_bytes(address[0:2]), 0)


from datetime import datetime
import tkinter as tk
from tkinter import *
from tkinter import ttk
import tkinter.scrolledtext as ScrolledText

from eltakobus.message import EltakoPoll, EltakoDiscoveryReply, EltakoDiscoveryRequest, EltakoMessage, prettify, Regular1BSMessage, EltakoWrapped1BS
from esp2_gateway_adapter.esp3_serial_com import ESP3SerialCommunicator

from eo_man import LOGGER

from ..data.data_helper import b2s, a2s
from ..data.data_manager import DataManager
from ..controller.app_bus import AppBus, AppBusEventType

class LogOutputPanel():

    def __init__(self, main: Tk, app_bus:AppBus, data_manager:DataManager):
        self.app_bus = app_bus
        self.data_manager = data_manager

        pane = ttk.Frame(main, padding=2, height=150)
        # pane.grid(row=2, column=0, sticky="nsew", columnspan=3)
        self.root = pane

        tool_bar = ttk.Frame(pane, height=24)
        tool_bar.pack(expand=False, fill=X, anchor=NW, pady=(0,4))

        # l = Label(tool_bar, text="Search")
        # l.pack(side=LEFT, padx=(2,0))

        # self.e_search = ttk.Entry(tool_bar, width="14") 
        # self.e_search.pack(side=LEFT, padx=(2,0))

        self.show_telegram_values = tk.BooleanVar(value=True)
        cb_show_values = ttk.Checkbutton(tool_bar, text="Show Telegram Values", variable=self.show_telegram_values)
        cb_show_values.pack(side=LEFT, padx=(2,0))

        self.show_esp2_binary =tk.BooleanVar(value=False)
        cb_show_es2_binary = ttk.Checkbutton(tool_bar, text="Show ESP2 Binary", variable=self.show_esp2_binary)
        cb_show_es2_binary.pack(side=LEFT, padx=(2,0))

        self.show_esp3_binary =tk.BooleanVar(value=False)
        cb_show_es3_binary = ttk.Checkbutton(tool_bar, text="Show ESP3 Binary", variable=self.show_esp3_binary)
        cb_show_es3_binary.pack(side=LEFT, padx=(2,0))

        self.st = ScrolledText.ScrolledText(pane, border=3,  height=150, 
                                            state=DISABLED, bg='black', fg='lightgrey', 
                                            font=('Arial', 14), padx=5, pady=5)
        self.st.configure(font='TkFixedFont')
        self.st.pack(expand=True, fill=BOTH)
        # self.st.grid(row=2, column=0, sticky="nsew", columnspan=3)

        app_bus.add_event_handler(AppBusEventType.SERIAL_CALLBACK, self.serial_callback)
        app_bus.add_event_handler(AppBusEventType.LOG_MESSAGE, self.receive_log_message)


    def serial_callback(self, data:dict):
        telegram:EltakoMessage = data['msg']
        current_base_id:str = data['base_id']

        # filter out poll messages
        filter = type(telegram) not in [EltakoPoll]
        # filter out empty telegrams (generated when sending telegrams with FAM-USB)
        try:
            filter &= (int.from_bytes(telegram.address, 'big') > 0 and int.from_bytes(telegram.payload, 'big'))
        except:
            pass

        if filter:
            tt = type(telegram).__name__
            adr = str(telegram.address) if isinstance(telegram.address, int) else b2s(telegram.address)
            if hasattr(telegram, 'reported_address'):
                adr = telegram.reported_address
            payload = ''
            if hasattr(telegram, 'data'):
                payload += ', data: '+b2s(telegram.data)
            elif hasattr(telegram, 'payload'):
                payload += ', payload: '+b2s(telegram.payload)
            
            if hasattr(telegram, 'status'):
                payload += ', status: '+ a2s(telegram.status, 1)

            values = ''
            eep, values = self.data_manager.get_values_from_message_to_string(telegram, current_base_id)
            if eep is not None: 
                if values is not None:
                    values = f" => values for EEP {eep.__name__}: ({values})"
                else:
                    values = f" => No matching value for EEP {eep.__name__}"
            else:
                values = ''

            display_values:str = values if self.show_telegram_values.get() else ''
            display_esp2:str = f", ESP2: {telegram.serialize().hex()}" if self.show_esp2_binary.get() else ''
            display_esp3:str = f", ESP3: { ''.join(f'{num:02x}' for num in ESP3SerialCommunicator.convert_esp2_to_esp3_message(telegram).build())}" if self.show_esp3_binary.get() else ''

            log_msg     = f"Received Telegram: {tt} from {adr}{payload}{values}"
            LOGGER.info(log_msg)

            display_msg = f"Received Telegram: {tt} from {adr}{payload}{display_values}{display_esp2}{display_esp3}"
            self.receive_log_message({'msg': display_msg, 'color': 'darkgrey'})


    def receive_log_message(self, data):
        msg = data.get('msg', False)
        if not msg: return

        # time_format = "%d.%b %Y %H:%M:%S"
        time_format = "%Y-%m-%d %H:%M:%S.%f"
        time_str = datetime.now().strftime(time_format)
        msg = f"{time_str}: {msg}"

        self.st.configure(state='normal')
        self.st.insert(tk.END, msg + '\n')
        self.st.configure(state='disabled')
        
        color = data.get('color', False)
        if color:
            final_index = str(self.st.index(tk.END))
            num_of_lines = int(final_index.split('.')[0])-2
            self.st.tag_config('mark_'+color, foreground=color)
            self.st.tag_add('mark_'+color, f"{num_of_lines}.0", f"{num_of_lines}.{len(msg)}")
        
        self.st.yview(tk.END)


import time
import threading
from esp2_gateway_adapter.esp3_tcp_com import TCP2SerialCommunicator, detect_lan_gateways

class NetworkGatewayDetector(threading.Thread):


    def __init__(self, callback_esp3_gw, callback_esp2_gw):
        self._running = False
        self.callback_esp3_gw = callback_esp3_gw
        self.callback_esp2_gw = callback_esp2_gw

    def stop(self):
        self._running = False
        self.join()

    def run(self):
        if not self._running:
            self._running = True

            while self._running:

                esp3_gw_list = detect_lan_gateways()
                for gw in esp3_gw_list:
                    self.callback_esp3_gw(gw)


                time.sleep(5)



import os
import tomli
import requests

class ApplicationInfo():

    app_info:dict[str:str]=None
    pypi_info_latest_versino:dict=None
    pypi_info_current_version:dict=None

    @classmethod
    def get_app_info(cls, filename:str=None):
        
        if not cls.app_info:

            app_info = {}
            parent_folder = os.path.join(os.path.dirname(__file__), '..')
            pyproject_file = os.path.join(parent_folder, 'pyproject.toml')
            if not os.path.isfile(pyproject_file):
                pyproject_file = os.path.join(parent_folder, '..', 'pyproject.toml')
                if not os.path.isfile(pyproject_file):
                    pyproject_file = None
                
            if pyproject_file:
                with open(pyproject_file, "rb") as f:
                    pyproject_data = tomli.load(f)
                app_info['version'] = pyproject_data["project"]["version"]
                app_info['name'] = pyproject_data["project"]["name"]
                app_info['author'] = ', '.join([ f"{a.get('name')} {a.get('email', '')}".strip() for a in pyproject_data["project"]["authors"]])
                app_info['home-page'] = pyproject_data["project"]["urls"]["Homepage"]
                app_info['license'] = pyproject_data["project"]["license"]["text"]

            if not filename:
                parent_folder = os.path.join(os.path.dirname(__file__), '..', '..')
                for f in os.listdir(parent_folder):
                    if os.path.isdir(os.path.join(parent_folder, f)):
                        # for installed package => get info from package metadata folder
                        if f.startswith('eo_man-') and f.endswith('.dist-info'):
                            filename = os.path.join(parent_folder, f, 'METADATA')
                            break
                        # for development environment => get info from built package (needs to be built first)
                        if 'eo_man.egg-info' in f:
                            filename = os.path.join(parent_folder, f, 'PKG-INFO')
                            break
            
            if filename and os.path.isfile(filename):
                with open(filename, 'r', encoding='utf-8') as file:
                    for l in file.readlines():
                        if l.startswith('Name:'):
                            app_info['name'] = l.split(':',1)[1].strip()
                        elif l.startswith('Version:'):
                            app_info['version'] = l.split(':',1)[1].strip()
                        elif l.startswith('Summary:'):
                            app_info['summary'] = l.split(':',1)[1].strip()
                        elif l.startswith('Home-page:'):
                            app_info['home-page'] = l.split(':',1)[1].strip()
                        elif l.startswith('Author:'):
                            app_info['author'] = l.split(':',1)[1].strip()
                        elif l.startswith('License:'):
                            app_info['license'] = l.split(':',1)[1].strip()
                        elif l.startswith('Requires-Python:'):
                            app_info['requires-python'] = l.split(':',1)[1].strip()

            
            # get latest version
            cls.pypi_info_latest_versino = cls._get_info_from_pypi()
            app_info['lastest_available_version'] = cls.pypi_info_latest_versino.get('info', {}).get('version', None)
            # get current/installed version
            if app_info['version'] is not None and app_info['version'] != '':
                cls.pypi_info_current_version = cls._get_info_from_pypi(app_info['version'])

            cls.app_info = app_info
            
        return cls.app_info
                        
    @classmethod
    def _get_info_from_pypi(cls, version:str=''):
        if version is not None and version != '':
            if not version.startswith('/') and not version.endswith('/'):
                version = version + '/'

        url = f"https://pypi.org/pypi/eo-man/{version}json"

        response = requests.get(url)
        if response.status_code == 200:
            return response.json()
            
        return {}


    @classmethod
    def get_app_info_as_str(cls, separator:str='\n', prefix:str='') -> str:
        result = ''
        for k, v in cls.get_app_info().items():
            result += f"{prefix}{k.title()}: {v}{separator}"
        return result

    @classmethod
    def get_package_name(cls) -> str:
        return cls.get_app_info().get('name', 'unknown')
    
    @classmethod
    def get_version(cls) -> str:
        return cls.get_app_info().get('version', 'unknown')

    @classmethod
    def get_summary(cls) -> str:
        return cls.get_app_info().get('summery', 'unknown')
    
    @classmethod
    def get_home_page(cls) -> str:
        return cls.get_app_info().get('home-page', 'unknown')
    
    @classmethod
    def get_author(cls) -> str:
        return cls.get_app_info().get('author', 'unknown')
    
    @classmethod
    def get_license(cls) -> str:
        return cls.get_app_info().get('license', 'unknown')

    @classmethod
    def get_requires_python(cls) -> str:
        return cls.get_app_info().get('requires-python', 'unknown')
    
    @classmethod
    def get_lastest_available_version(cls) -> str:
        return cls.get_app_info().get('lastest_available_version', 'unknown')

    @classmethod
    def is_version_up_to_date(cls) -> bool:
        cv = cls.get_app_info().get('version', '0.0.0') 
        lv = cls.pypi_info_latest_versino['info']['version']

        for i in range(0,3):
            l = int(lv.split('.')[i])
            c = int(cv.split('.')[i])
            if l > c:
                return False
            elif c > l:
                return True
        return True


from functools import reduce

from tkinter import *
import tkinter as tk
from tkinter import Tk, ttk

from eltakobus.message import *
from eltakobus.eep import A5_08_01
from eltakobus.util import AddressExpression

from ..controller.app_bus import AppBus, AppBusEventType
from ..controller.serial_controller import SerialController
from ..data import data_helper
from ..data.data_manager import DataManager
from ..data.message_history import MessageHistoryEntry


class SendMessageWindow():

    ADDRESS_COLOR = "green"
    DATA_COLOR = "red"
    STATUS_COLOR = "blue"
    ORG_COLOR = "orange"

    def __init__(self, main:Tk, app_bus:AppBus, data_manager:DataManager, serial_controller:SerialController):
        self.main = main
        self.app_bus = app_bus
        self.data_manager = data_manager
        self.serial_controller = serial_controller

        self.app_bus.add_event_handler(AppBusEventType.CONNECTION_STATUS_CHANGE, self.on_connection_state_changed)

        self.popup = None

        self.data = {0: '01', 1: '00', 2:'00', 3:'09'}
        self.address = {0: '00', 1: '00', 2:'00', 3:'00'}
        self.status = '00'

        self.entered_keys = ''
        self.current_telegram = None

    def show_window(self):
        if self.popup:
            self.center_window()
            return
        
        self.popup = Toplevel(self.main, padx=4, pady=4)
        self.popup.wm_title("Send Message")

        self.col_widths = [90, 204, 8, 360, 28]
        for idx, w in enumerate(self.col_widths):
            self.popup.columnconfigure(idx, minsize=w)
        
        for idx in range(0,8):
            self.popup.rowconfigure(idx, weight=0)
        

        _pady = (4,0)
        _padx = (2,0)

        row = 0
        l = ttk.Label(self.popup, text="Message Type: ", foreground=self.ORG_COLOR)
        l.grid(row=row, column=0, sticky=W)

        self.l_msg_type = ttk.Label(self.popup, text='')
        self.l_msg_type.grid(row=row, column=1, sticky=W, pady=_pady, padx=_padx)
        self.cb_msg_type = ttk.Combobox(self.popup, state="readonly", width="18") 
        self.cb_msg_type['values'] = ['RPS (Org = 0x05)', '1BS (Org = 0x06)', '4BS (Org = 0x07)']
        self.cb_msg_type.set('4BS (Org = 0x07)')
        self.cb_msg_type.grid(row=row, column=1, sticky=W, pady=_pady, padx=_padx)
        self.cb_msg_type.bind('<<ComboboxSelected>>', self.on_message_type_changed)


        row += 1
        l = ttk.Label(self.popup, text="Data: ", foreground=self.DATA_COLOR)
        l.grid(row=row, column=0, sticky=W, pady=_pady, padx=_padx)

        # data fields
        f = ttk.Frame(self.popup)
        f.grid(row=row, column=1, sticky=W, pady=_pady, padx=_padx)
        self.cb_data_0 = ttk.Combobox(f, state="readonly", width="3")
        self.cb_data_1 = ttk.Combobox(f, state="readonly", width="3")
        self.cb_data_2 = ttk.Combobox(f, state="readonly", width="3")
        self.cb_data_3 = ttk.Combobox(f, state="readonly", width="3")
        for idx, cb_d in enumerate([self.cb_data_0, self.cb_data_1, self.cb_data_2, self.cb_data_3]):
            cb_d['values'] = [data_helper.a2s(i,1) for i in range(0,256)]
            cb_d.set(self.data[idx])
            cb_d.bind('<<ComboboxSelected>>', self.show_message )
            cb_d.bind('<FocusIn>', self.on_focus_combobox)
            cb_d.bind('<Key>', self.select_by_entered_keys)
            cb_d.bind('<Return>', lambda e: self.show_message(e, True))
            cb_d.pack(side=LEFT)

        
        row += 1
        l = ttk.Label(self.popup, text="Sender Id: ", foreground=self.ADDRESS_COLOR)
        l.grid(row=row, column=0, sticky=W, pady=_pady, padx=_padx)
        f = ttk.Frame(self.popup)
        f.grid(row=row, column=1, sticky=W, pady=_pady, padx=_padx)

        # address field
        self.cb_sender_id_0 = ttk.Combobox(f, state="readonly", width="3")
        self.cb_sender_id_1 = ttk.Combobox(f, state="readonly", width="3")
        self.cb_sender_id_2 = ttk.Combobox(f, state="readonly", width="3")
        self.cb_sender_id_3 = ttk.Combobox(f, state="readonly", width="3")
        for cb_s in [self.cb_sender_id_0, self.cb_sender_id_1, self.cb_sender_id_2, self.cb_sender_id_3]:
            cb_s['values'] = [data_helper.a2s(i,1) for i in range(0,256)]
            cb_s.bind('<<ComboboxSelected>>', self.show_message )
            cb_s.bind('<FocusIn>', self.on_focus_combobox)
            cb_s.bind('<Key>', self.select_by_entered_keys)
            cb_s.bind('<Return>', lambda e: self.show_message(e, True))
            cb_s.pack(side=LEFT)
        self.set_sender_ids()
        
        # status
        row += 1
        l = ttk.Label(self.popup, text="Status: ", foreground=self.STATUS_COLOR)
        l.grid(row=row, column=0, sticky=W, pady=_pady, padx=_padx)

        self.cb_status = ttk.Combobox(self.popup, width="3", state="readonly")
        self.cb_status['values'] = [data_helper.a2s(i,1) for i in range(0,256)]
        self.cb_status.set(self.status)
        self.cb_status.bind('<<ComboboxSelected>>', self.show_message )
        self.cb_status.bind('<FocusIn>', self.on_focus_combobox)
        self.cb_status.bind('<Key>', self.select_by_entered_keys)
        self.cb_status.bind('<Return>', lambda e: self.show_message(e, True))
        self.cb_status.grid(row=row, column=1, sticky=W, pady=_pady, padx=_padx)


        # out going message
        row += 1
        self.out_going_msg_boolvar = BooleanVar()
        self.out_going_msg_boolvar.set(True)
        cb = ttk.Checkbutton(self.popup, text="Out-Going Telegram", command=self.show_message, variable=self.out_going_msg_boolvar)
        cb.grid(row=row, sticky=W, column=1, pady=_pady, padx=_padx)


        row += 1
        l = ttk.Label(self.popup, text="")
        l.grid(row=row, column=0, sticky=W, columnspan=2, pady=_pady, padx=_padx)


        row += 1
        l = ttk.Label(self.popup, text="Telegram:")
        l.grid(row=row, column=0, sticky=W, columnspan=2, pady=_pady, padx=_padx)


        row += 1
        self.popup.rowconfigure(row, weight=1)
        self.f_telegram = ttk.Frame(self.popup)
        self.f_telegram.grid(row=row, column=0, sticky=NSEW, columnspan=2, pady=_pady, padx=_padx)


        row += 1
        l = ttk.Label(self.popup, text="")
        l.grid(row=row, column=0, sticky=W, columnspan=2, pady=_pady, padx=_padx)


        # close button
        row += 1
        b_close = ttk.Button(self.popup, text="Close", command=self.close)
        b_close.bind('<Return>', lambda e: self.close())
        b_close.grid(row=row, column=0, sticky=EW)

        # send button
        self.b_send = ttk.Button(self.popup, text="Send", command=lambda send=True: self.show_message(None, send))
        if not self.serial_controller.is_serial_connection_active():
            self.b_send.config(state=DISABLED)
        self.b_send.bind('<Return>', lambda event, send=True: self.show_message(event, send))
        self.b_send.grid(row=row, column=1, sticky=EW, padx=(0,8))


        ttk.Separator(self.popup, orient=VERTICAL).grid(column=2, row=0, rowspan=10, sticky=NSEW)

        # favourites

        row = 0
        l = ttk.Label(self.popup, text="Favourites")
        l.grid(row=row, column=3, sticky=W)

        row += 1
        rowspan = 7
        f = ttk.Frame(self.popup, width=self.col_widths[3])
        f.grid(row=row, rowspan=rowspan, column=3, sticky=NSEW)
        s = ttk.Style()
        s.configure('SendMessageWindow.Treeview', rowheight=38)
        scrollbar = ttk.Scrollbar(f)
        self.lb_fav_msg = ttk.Treeview(f, yscrollcommand=scrollbar.set, show="tree", height=4, selectmode='browse', style='SendMessageWindow.Treeview')
        scrollbar.configure(command=self.lb_fav_msg.yview)
        scrollbar.pack(side=RIGHT, fill=Y)
        self.lb_fav_msg.pack(side=LEFT, fill=BOTH, expand=True)    

        if self.data_manager.send_message_template_list:
            for history_entry in self.data_manager.send_message_template_list:
                msg = ESP2Message.parse(history_entry.message)
                org = msg.body[1]
                data = msg.body[2:6]
                address = msg.body[6:10]
                status = msg.body[10]
                outgoing = {(3 << 5) + 11: True, (0 << 5) + 11: False}[msg.body[0]]

                if org == 7:    msg = Regular4BSMessage(address, status, data, outgoing)
                elif org == 6:  msg = Regular1BSMessage(address, status, data[0:1], outgoing)
                elif org == 5:  msg = RPSMessage(address, status, data[0:1], outgoing)

                self.add_fav_msg(msg, history_entry.name)


        row += rowspan
        f = ttk.Frame(self.popup)
        f.grid(row=row, column=3, sticky=E+W)
        self.b_add_fav = ttk.Button(f, text="Add as", command=lambda: self.add_fav_msg(self.current_telegram) )
        self.b_add_fav.bind('<Return>', lambda e: self.add_fav_msg(self.current_telegram))
        self.b_add_fav.pack(side=LEFT, pady=(4,2), padx=(0,2))
        self.e_name = ttk.Entry(f)
        self.e_name.bind('<Return>', lambda e: self.add_fav_msg(self.current_telegram))
        self.e_name.pack(fill=X, pady=(4,2), padx=(0,16))

        row += 1
        f = ttk.Frame(self.popup)
        f.grid(row=row, column=3, sticky=E+W)

        self.b_remove_fav = ttk.Button(f, text="Remove", command=self.remove_fav_msg)
        self.b_remove_fav.bind('<Return>', lambda e: self.remove_fav_msg())
        self.b_remove_fav.pack(side=LEFT, pady=2, padx=(0,2))
        self.b_apply_fav = ttk.Button(f, text="Apply", command=self.apply_fav_msg)
        self.b_apply_fav.bind('<Return>', lambda e: self.apply_fav_msg())
        self.b_apply_fav.pack(fill=X, pady=2, padx=(0,16))


        self.show_message(None)

        # self.popup.wm_attributes('-toolwindow', 'True')
        self.popup.protocol("WM_DELETE_WINDOW", self.close)
        self.popup.resizable (width=False, height=False)
        
        self.popup.after(10, self.center_window)

        self.main.wait_window(self.popup)


    def remove_fav_msg(self):
        if self.lb_fav_msg.selection():
            for item in self.lb_fav_msg.selection():
                self.lb_fav_msg.delete(item)

        self.send_update_of_current_message_templates()
        

    def send_update_of_current_message_templates(self):
        entries = list()
        for c_id in self.lb_fav_msg.get_children():
           
            name = self.lb_fav_msg.item(c_id)['text'].split('\n')[0]
            msg = bytes.fromhex(c_id)

            entries.append(MessageHistoryEntry(name, msg))

        self.app_bus.fire_event(AppBusEventType.SEND_MESSAGE_TEMPLATE_LIST_UPDATED, entries)


    def add_fav_msg(self, msg:EltakoMessage,name:str=''):
        if msg:
            id = data_helper.b2s(msg.serialize(), separator='')

            text = ''
            if isinstance(msg, Regular4BSMessage): text += "4BS: "
            elif isinstance(msg, Regular1BSMessage): text += "1BS: "
            elif isinstance(msg, RPSMessage): text += "RPS: "

            text += f"Data: {data_helper.b2s(msg.data, ' ')}, "
            text += f"Address: {data_helper.b2s(msg.address, ' ')}, "
            text += "Status: {:02X}".format(msg.status)

            if not name and self.e_name.get():
                name = self.e_name.get()
                self.e_name.delete(0, END)

            text = f"{name}\n{text}"

            if not self.lb_fav_msg.exists(id):
                self.lb_fav_msg.insert(parent="", index="end", iid=id, text=text)
            else:
                self.lb_fav_msg.item(id, text=text)

            self.send_update_of_current_message_templates()

    def apply_fav_msg(self):
        if self.lb_fav_msg.selection():
            msg = ESP2Message.parse(bytes.fromhex(self.lb_fav_msg.selection()[0]))
            data = msg.body[2:6]
            address = msg.body[6:10]
            status = msg.body[10]
            outgoing = {(3 << 5) + 11: True, (0 << 5) + 11: False}[msg.body[0]]

            item_text = self.lb_fav_msg.item(self.lb_fav_msg.selection()[0])['text']
            if '4BS' in item_text:   self.cb_msg_type.set('4BS (Org = 0x07)')
            elif '1BS' in item_text: self.cb_msg_type.set('1BS (Org = 0x06)')
            elif 'RPS' in item_text: self.cb_msg_type.set('RPS (Org = 0x05)')
            self.on_message_type_changed()

            self.cb_data_0.set( "{:02X}".format(data[0]) )
            if '4BS' in item_text:
                self.cb_data_1.set( "{:02X}".format(data[1]) )
                self.cb_data_2.set( "{:02X}".format(data[2]) )
                self.cb_data_3.set( "{:02X}".format(data[3]) )

            self.cb_sender_id_3.set( "{:02X}".format(address[3]) )
            if str(self.cb_sender_id_0['state']) not in ['disabled']:
                self.cb_sender_id_0.set( "{:02X}".format(address[0]) )
                self.cb_sender_id_1.set( "{:02X}".format(address[1]) )
                self.cb_sender_id_2.set( "{:02X}".format(address[2]) )

            self.cb_status.set( "{:02X}".format(status) )
            self.out_going_msg_boolvar.set(outgoing)

            self.show_message()


    def on_focus_combobox(self, event=None):
        self.entered_keys = ''


    def select_by_entered_keys(self, event=None):

        if event.char.upper() not in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']:
            return

        if event is not None:
            if len(self.entered_keys) < 2:
                self.entered_keys += event.char.upper()
            else:
                self.entered_keys = event.char.upper()

        cb:ttk.Combobox = self.popup.focus_get()
        if len(self.entered_keys) == 2:
            cb.set(self.entered_keys)
        elif len(self.entered_keys) == 1:
            for v in cb['values']:
                if v[0] == self.entered_keys:
                    cb.set(v)
                    break

        self.show_message()


    def on_message_type_changed(self, event=None):
        msg_type = self.cb_msg_type.get()
        for idx, cb_d in enumerate([self.cb_data_1, self.cb_data_2, self.cb_data_3]):
            if '4BS' not in msg_type:
                if cb_d.get() != '': self.data[idx+1] = cb_d.get()
                cb_d.config(state=DISABLED)
                cb_d.set('')
            else:
                cb_d.config(state="readonly")
                cb_d.set(self.data[idx+1])

        self.show_message(event)


    def center_window(self):
        if self.popup:
            w = reduce(lambda x, y: x + y, self.col_widths) # popup.winfo_width()
            h = self.popup.winfo_height()
            x = self.main.winfo_x() + self.main.winfo_width()/2 - w/2
            y = self.main.winfo_y() + self.main.winfo_height()/2 - h/2
            self.popup.geometry('%dx%d+%d+%d' % (w, h, x, y))
            self.popup.focus()
            self.popup.state(newstate=NORMAL)


    def close(self):
        if self.popup:
            self.popup.destroy()
            self.popup = None
    
    
    def set_sender_ids(self):
        # not wired gateway
        if not self.serial_controller.is_connected_gateway_device_bus():
            base_id = self.serial_controller.current_base_id
        else:
            base_id = '00-00-B0-00'
        
        if self.serial_controller.current_base_id and not self.serial_controller.is_connected_gateway_device_bus():
            base_id = self.serial_controller.current_base_id
            self.cb_sender_id_0.config(state="readonly")
            self.cb_sender_id_1.config(state="readonly")
            self.cb_sender_id_2.config(state="readonly")
            self.cb_sender_id_0.set( base_id.split('-')[0] )
            self.cb_sender_id_1.set( base_id.split('-')[1] )
            self.cb_sender_id_2.set( base_id.split('-')[2] )
            self.cb_sender_id_3.set(self.address[3])
            self.cb_sender_id_0.config(state=DISABLED)
            self.cb_sender_id_1.config(state=DISABLED)
            self.cb_sender_id_2.config(state=DISABLED)

            self.cb_sender_id_3['values'] = [data_helper.a2s(i+int(base_id.split('-')[3]) ,1) for i in range(0,128)]
            self.cb_sender_id_3.set(base_id.split('-')[3])

        else:
            self.cb_sender_id_0.config(state="readonly")
            self.cb_sender_id_1.config(state="readonly")
            self.cb_sender_id_2.config(state="readonly")

            self.cb_sender_id_3['values'] = [data_helper.a2s(i,1) for i in range(0,256)]
            
            self.cb_sender_id_0.set(self.address[0])
            self.cb_sender_id_1.set(self.address[1])
            self.cb_sender_id_2.set(self.address[2])
            self.cb_sender_id_3.set(self.address[3])


    def on_connection_state_changed(self, data):
        if self.popup:
            if data.get('connected'):
                self.b_send.config(state=NORMAL)
            else:
                self.b_send.config(state=DISABLED)
            self.set_sender_ids()


    def show_message(self, event=None, send:bool=False):
        msg_type = self.cb_msg_type.get()
        try:
            self.address[0] = self.cb_sender_id_0.get()
            self.address[1] = self.cb_sender_id_1.get()
            self.address[2] = self.cb_sender_id_2.get()
            self.address[3] = self.cb_sender_id_3.get()
            sender_id_str = '-'.join(self.address.values())
            sender_id = AddressExpression.parse( sender_id_str )[0]
            data_str = f"{self.cb_data_0.get()}-{self.cb_data_1.get()}-{self.cb_data_2.get()}-{self.cb_data_3.get()}".replace('---', '-00-00-00')
            data = AddressExpression.parse( data_str )[0]
            self.status = self.cb_status.get()
            status = int(self.cb_status.get(), 16)
            
            if '4BS' in msg_type:
                msg = Regular4BSMessage(sender_id, status, data, self.out_going_msg_boolvar.get())
            elif '1BS' in msg_type:
                data = data[0:1]
                msg = Regular1BSMessage(sender_id, status, data, self.out_going_msg_boolvar.get())
            elif 'RPS' in msg_type:
                data = data[0:1]
                msg = RPSMessage(sender_id, status, data, self.out_going_msg_boolvar.get())

            self.current_telegram = msg

            msg_text = b2a(msg.serialize()).upper()
            if '4BS' not in msg_type: msg_text = msg_text[0:15] + msg_text[24:]
            self.show_telegram_text(msg_text, msg_type)

            if send and self.serial_controller.is_serial_connection_active():
                self.serial_controller.send_message(msg)
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Send Message: {str(msg)}", 'color':'green', 'log-level':'INFO'})
        except Exception as e:
            self.show_telegram_text('Invalid telegram!')
            self.current_telegram = None


    def show_telegram_text(self, text:str, msg_type:str=''):
        for widgets in self.f_telegram.winfo_children():
            widgets.destroy()

        if 'invalid' in text.lower():
            l = ttk.Label(self.f_telegram, text=text, foreground='red')
            l.pack(side=LEFT, anchor=NW, padx=(12,0))
        else:
            l = ttk.Label(self.f_telegram, text=text[0:9], foreground='darkgrey')
            l.pack(side=LEFT, anchor=NW, padx=(12,0))

            # org
            l = ttk.Label(self.f_telegram, text=text[9:12], foreground=self.ORG_COLOR, font = ('Sans','10','bold'))
            l.pack(side=LEFT, anchor=NW)

            # data
            data_end = 24
            if '4BS' not in msg_type: data_end = 12 + 3
            l = ttk.Label(self.f_telegram, text=text[12:data_end], foreground=self.DATA_COLOR)
            l.pack(side=LEFT, anchor=NW)

            # sender id
            sender_end = data_end + 12
            l = ttk.Label(self.f_telegram, text=text[data_end:sender_end], foreground=self.ADDRESS_COLOR)
            l.pack(side=LEFT, anchor=NW)

            # status
            l = ttk.Label(self.f_telegram, text=text[sender_end:sender_end+3], foreground=self.STATUS_COLOR)
            l.pack(side=LEFT, anchor=NW)

            # check sum
            l = ttk.Label(self.f_telegram, text=text[sender_end+3:sender_end+6], foreground='darkgrey')
            l.pack(side=LEFT, anchor=NW)


import threading
import time
from tkinter import *
from tkinter import ttk

from ..controller.app_bus import AppBus, AppBusEventType
from ..data.const import *
from ..data.homeassistant.const import CONF_ID, CONF_NAME
from ..data.filter import DataFilter
from ..data.data_manager import DataManager, Device
from ..data import data_helper
from ..icons.image_gallary import ImageGallery

from eltakobus.util import b2s
from eltakobus.message import EltakoMessage, RPSMessage, Regular1BSMessage, Regular4BSMessage, EltakoWrappedRPS


class DeviceTable():

    ICON_SIZE = (20,20)
    NON_BUS_DEVICE_LABEL:str="Distributed Devices"

    def __init__(self, main: Tk, app_bus:AppBus, data_manager:DataManager):
        
        self.blinking_enabled = True
        self.pane = ttk.Frame(main, padding=2)
        # self.pane.grid(row=0, column=0, sticky=W+E+N+S)
        self.root = self.pane

        # Scrollbar
        yscrollbar = ttk.Scrollbar(self.pane, orient=VERTICAL)
        yscrollbar.pack(side=RIGHT, fill=Y)

        xscrollbar = ttk.Scrollbar(self.pane, orient=HORIZONTAL)
        xscrollbar.pack(side=BOTTOM, fill=X)

        # Treeview
        columns = ("Address", "External Address", "Device Type", "Key Function", "Comment", "Export to HA Config", "HA Platform", "Device EEP", "Sender Address", "Sender EEP")
        self.treeview = ttk.Treeview(
            self.pane,
            show="tree headings", 
            selectmode="browse",
            yscrollcommand=yscrollbar.set,
            xscrollcommand=xscrollbar.set,
            columns=(0,1,2,3,4,5,6,7,8,9),
        )
        self.treeview.pack(expand=True, fill=BOTH)
        yscrollbar.config(command=self.treeview.yview)
        xscrollbar.config(command=self.treeview.xview)

        def sort_rows_in_treeview(tree:ttk.Treeview, col_i:int, descending:bool, partent:str=''):
            data = [(tree.set(item, col_i), item) for item in tree.get_children(partent)]
            data.sort(reverse=descending)
            for index, (val, item) in enumerate(data):
                tree.move(item, partent, index)
            
            for item in tree.get_children(partent):
                sort_rows_in_treeview(tree, col_i, descending, item)

        def sort_treeview(tree:ttk.Treeview, col:int, descending:bool):
            i = columns.index(col)
            for item in tree.get_children(''):
                sort_rows_in_treeview(tree, i, descending, item)
            tree.heading(i, command=lambda c=col, d=(not descending): sort_treeview(tree, c, d))

        self.treeview.column('#0', anchor="w", width=250, minwidth=250)#, stretch=NO)
        for col in columns:
            # Treeview headings
            i = columns.index(col)
            if col in ['Key Function']:
                self.treeview.column(i, anchor="w", width=250, minwidth=250)#, stretch=NO)
            else:
                self.treeview.column(i, anchor="w", width=80, minwidth=80)#, stretch=NO)
            self.treeview.heading(i, text=col, anchor="center", command=lambda c=col, d=False: sort_treeview(self.treeview, c, d))
        
        # self.menu = Menu(main, tearoff=0)
        # self.menu.add_command(label="Cut")
        # self.menu.add_command(label="Copy")
        # self.menu.add_command(label="Paste")
        # self.menu.add_command(label="Reload")
        # self.menu.add_separator()
        # self.menu.add_command(label="Rename")

        self.treeview.tag_configure('related_devices')
        self.treeview.tag_configure('blinking', background='lightblue')

        # self.treeview.bind('<ButtonRelease-1>', self.on_selected)
        self.treeview.bind('<<TreeviewSelect>>', self.on_selected)
        # self.treeview.bind("<Button-3>", self.show_context_menu)

        self.check_if_wireless_network_exists()

        self.current_data_filter:DataFilter = None
        self.app_bus = app_bus
        self.app_bus.add_event_handler(AppBusEventType.DEVICE_SCAN_STATUS, self.device_scan_status_handler)
        self.app_bus.add_event_handler(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, self.update_device_representation_handler)
        self.app_bus.add_event_handler(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, self.update_sensor_representation_handler)
        self.app_bus.add_event_handler(AppBusEventType.LOAD_FILE, self._reset)
        self.app_bus.add_event_handler(AppBusEventType.SET_DATA_TABLE_FILTER, self._set_data_filter_handler)
        self.app_bus.add_event_handler(AppBusEventType.SERIAL_CALLBACK, self._serial_callback_handler)

        self.data_manager = data_manager

        # initial loading
        if self.data_manager.selected_data_filter_name is not None:
            self._set_data_filter_handler(self.data_manager.data_fitlers[self.data_manager.selected_data_filter_name])
        for d in self.data_manager.devices.values():
            parent = self.NON_BUS_DEVICE_LABEL if not d.is_bus_device() else None
            self.update_device_handler(d, parent)


    def _set_data_filter_handler(self, filter):
        self.current_data_filter = filter

        self._reset(None)
        for d in self.data_manager.devices.values():
            if d.bus_device:
                self.update_device_handler(d)
            else:
                self.update_device_handler(d, parent=self.NON_BUS_DEVICE_LABEL)


    def _reset(self, data):
        for item in self.treeview.get_children():
            self.treeview.delete(item)
        self.check_if_wireless_network_exists()


    def on_selected(self, event):
        device_external_id = self.treeview.focus()
        device = self.data_manager.get_device_by_id(device_external_id)
        if device is not None:
            self.app_bus.fire_event(AppBusEventType.SELECTED_DEVICE, device)

        self.mark_related_elements(device_external_id)


    def mark_related_elements(self, device_external_id:str) -> None:
        for iid in self.treeview.tag_has( 'related_devices' ):
            self.treeview.item( iid, tags=() )

        devices = self.data_manager.get_related_devices(device_external_id)
        for d in devices:
            if self.treeview.exists(d.external_id):
                self.treeview.item(d.external_id, tags=('related_devices'))


    def show_context_menu(self, event):
        try:
            self.menu.tk_popup(event.x_root, event.y_root)
        finally:
            self.menu.grab_release()


    def insert_device(self, device:Device):
        v=("", b2s(device.address[0]), "", "")
        self.treeview.insert(parent="", index="end", text=device.id_string, values=v)

        
    def device_scan_status_handler(self, status:str):
        if status in ['STARTED']:
            #TODO: disable treeview or menue of it
            # self.treeview.config(state=DISABLED)
            pass
        elif status in ['FINISHED']:
            #TODO: enable treeview or menue of it
            # self.treeview.config(state=NORMAL)
            pass


    def add_fam14(self, d:Device):
        if d.is_fam14():
            if not self.treeview.exists(d.base_id):
                text = ""
                comment = ""
                text = d.name
                comment = d.comment if d.comment is not None else "" 
                in_ha = d.use_in_ha
                self.treeview.insert(parent="", 
                                     index=0, 
                                     iid=d.external_id, 
                                     text=" " + text, 
                                     values=("", "", "", "", comment, in_ha, "", "", ""),
                                     image=ImageGallery.get_fam14_icon(self.ICON_SIZE),
                                     open=True)
            else:
                self.treeview.item(d.base_id, 
                                   text=" " + d.name, 
                                   values=("", "", "", "", d.comment, d.use_in_ha, "", "", ""),
                                   image=ImageGallery.get_fam14_icon(self.ICON_SIZE), 
                                   open=True)


    def check_if_wireless_network_exists(self):
        id = self.NON_BUS_DEVICE_LABEL
        if not self.treeview.exists(id):
            self.treeview.insert(parent="", 
                                 index="end", 
                                 iid=id, 
                                 text=" " + self.NON_BUS_DEVICE_LABEL, 
                                 values=("", "", "", "", "", "", "", "", ""), 
                                 image=ImageGallery.get_wireless_icon(self.ICON_SIZE),
                                 open=True)


    def update_device_representation_handler(self, d:Device):
        self.update_device_handler(d)


    def update_device_handler(self, d:Device, parent:str=None):

        if self.current_data_filter is not None and not self.current_data_filter.filter_device(d):
            return

        if not d.is_fam14():
            in_ha = d.use_in_ha
            ha_pl = "" if d.ha_platform is None else d.ha_platform
            eep = "" if d.eep is None else d.eep
            device_type = "" if d.device_type is None else d.device_type
            key_func = "" if d.key_function is None else d.key_function
            comment = "" if d.comment is None else d.comment
            sender_adr = "" if 'sender' not in d.additional_fields else d.additional_fields['sender'][CONF_ID]
            sender_eep = "" if 'sender' not in d.additional_fields else d.additional_fields['sender'][CONF_EEP]
            
            if d.is_usb300():
                image = ImageGallery.get_usb300_icon(self.ICON_SIZE)
            elif d.is_fam_usb():
                image = ImageGallery.get_fam_usb_icon(self.ICON_SIZE)
            elif d.is_fgw14_usb():
                image = ImageGallery.get_fgw14_usb_icon(self.ICON_SIZE)
            elif d.is_ftd14():
                image = ImageGallery.get_ftd14_icon(self.ICON_SIZE)
            elif d.is_EUL_Wifi_gw():
                image = ImageGallery.get_eul_gateway_icon(self.ICON_SIZE)
            elif d.is_mgw():
                image = ImageGallery.get_mgw_piotek_icon(self.ICON_SIZE)
            else:
                image = ImageGallery.get_blank(self.ICON_SIZE)

            _parent = d.base_id if parent is None else parent
            if not self.treeview.exists(_parent): self.add_fam14(self.data_manager.devices[_parent])
            if not self.treeview.exists(d.external_id):
                self.treeview.insert(parent=_parent, 
                                     index="end", 
                                     iid=d.external_id, 
                                     text=" " + d.name, 
                                     values=(d.address, d.external_id, device_type, key_func, comment, in_ha, ha_pl, eep, sender_adr, sender_eep), 
                                     open=True)
                self.treeview.item(d.external_id, image=image)
            else:
                # update device
                self.treeview.item(d.external_id, 
                                   text=" " + d.name, 
                                   values=(d.address, d.external_id, device_type, key_func, comment, in_ha, ha_pl, eep, sender_adr, sender_eep), 
                                   image=image,
                                   open=True)
                if self.treeview.parent(d.external_id) != _parent:
                    self.treeview.move(d.external_id, _parent, 0)
        else:
            self.add_fam14(d)
        # self.trigger_blinking(d.external_id)
            

    def _serial_callback_handler(self, data:dict):
        message:EltakoMessage = data['msg']
        current_base_id:str = data['base_id']

        if type(message) in [RPSMessage, Regular1BSMessage, Regular4BSMessage, EltakoWrappedRPS]:
            if isinstance(message.address, int):
                adr = data_helper.a2s(message.address)
            else: 
                adr = b2s(message.address)

            if not adr.startswith('00-00-00-'):
                self.trigger_blinking(adr)
            elif current_base_id is not None:
                d:Device = self.data_manager.find_device_by_local_address(adr, current_base_id)
                if d is not None:
                    self.trigger_blinking(d.external_id)


    def trigger_blinking(self, external_id:str):
        if not self.blinking_enabled:
            return
        
        def blink(ext_id:str):
            for i in range(0,2):
                if self.treeview.exists(ext_id):
                    tags = self.treeview.item(ext_id)['tags']
                    if 'blinking' in tags:
                        if isinstance(tags, str):
                            self.treeview.item(ext_id, tags=() )
                        else:
                            tags.remove('blinking')
                            self.treeview.item(ext_id, tags=tags )
                    else:
                        if isinstance(tags, str):
                            self.treeview.item(ext_id, tags=('blinking') )
                        else:
                            tags.append('blinking')
                            self.treeview.item(ext_id, tags=tags )
                    time.sleep(.5)

            if self.treeview.exists(ext_id):
                tags = self.treeview.item(ext_id)['tags']
                if 'blinking' in tags:
                    if isinstance(tags, str):
                        self.treeview.item(ext_id, tags=() )
                    else:
                        tags.remove('blinking')
                        self.treeview.item(ext_id, tags=tags )

        t = threading.Thread(target=lambda ext_id=external_id: blink(ext_id))
        t.start()


    def update_sensor_representation_handler(self, d:Device):
        self.update_device_handler(d, parent=self.NON_BUS_DEVICE_LABEL)


from enum import Enum
from serial import rs485
from typing import Iterator
from termcolor import colored
import logging
import threading
import socket

import serial.tools.list_ports

from eltakobus import *
from eltakobus.device import *
from eltakobus.locking import buslocked, UNLOCKED
from eltakobus.message import Regular4BSMessage

from esp2_gateway_adapter.esp3_serial_com import ESP3SerialCommunicator
from esp2_gateway_adapter.esp3_tcp_com import TCP2SerialCommunicator, detect_lan_gateways
from esp2_gateway_adapter.esp2_tcp_com import ESP2TCP2SerialCommunicator

from ..data.const import GatewayDeviceType, get_gateway_type_by_name
from ..data import data_helper
from ..data.const import GatewayDeviceType as GDT, GATEWAY_DISPLAY_NAMES as GDN

from .gateway_registry import GatewayRegistry

from .app_bus import AppBusEventType, AppBus

class SerialController():

    USB_VENDOR_ID = 0x0403

    def __init__(self, app_bus:AppBus, gw_registry: GatewayRegistry) -> None:
        self.app_bus = app_bus
        self._serial_bus = None
        self.connected_gateway_type = None
        self.connected_mdns_service = ''
        self.current_base_id:str = None
        self.current_device_type:GatewayDeviceType = None
        self.gateway_id:str = None

        self.gw_registry:GatewayRegistry = gw_registry
        
        self.received_bus_device_discovery:Dict[str,List[EltakoDiscoveryReply]] = {}
        self.current_discovery_reply = None
        self.received_bus_device_memory:Dict[str,List[EltakoMemoryResponse]] = {}

        self.app_bus.add_event_handler(AppBusEventType.WINDOW_CLOSED, self.on_window_closed)
    
    
    
    def on_window_closed(self, data) -> None:
        self.kill_serial_connection_before_exit()


    def is_connected_gateway_device_bus(self):
        return self.connected_gateway_type.startswith('FAM14') or self.connected_gateway_type.startswith('FGW14-USB')

    
    def is_serial_connection_active(self) -> bool:
        return self._serial_bus is not None and self._serial_bus.is_active()


    def is_fam14_connection_active(self) -> bool:
        return self.is_serial_connection_active() and self._serial_bus.suppress_echo


    def _received_serial_event(self, message: ESP2Message) -> None:
        try:
            self.process_base_id_and_version_info(message)

            self.process_discovery_message(message)

            # log received message
            self.app_bus.fire_event(AppBusEventType.SERIAL_CALLBACK, {'msg': message, 
                                                                    'base_id': self.current_base_id,
                                                                    'gateway_id': self.gateway_id})
            
        except Exception as e:
            logging.exception(e)
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': e, 'log-level': 'ERROR', 'color': 'red'})
            

    def process_base_id_and_version_info(self, message: ESP2Message):
        # receive base id
        if message.body[:2] == b'\x8b\x98':
            self.current_base_id = b2s(message.body[2:6])
            device_type_index = message.body[6]
            # for virtual network gateway check which gateway is really sending
            if device_type_index > 0:
                gw_type = GDT.get_by_index(device_type_index-1)
                self.current_device_type = gw_type

            data = {
                'type': self.current_device_type, 
                'mdns_service': self.connected_mdns_service,
                'base_id': self.current_base_id, 
                'gateway_id': self.current_base_id,
                'tcm_version': '', 
                'api_version': ''
            }

            if hasattr(self._serial_bus, 'host') and hasattr(self._serial_bus, 'port'):
                data['address'] = f"{self._serial_bus._host}:{self._serial_bus._port}"

            self.app_bus.fire_event(AppBusEventType.ASYNC_TRANSCEIVER_DETECTED, data)

        # receive software version 
        if message.body[:2] == b'\x8b\x8c':
            tcm_sw_v = '.'.join(str(n) for n in message.body[2:6])
            api_v = '.'.join(str(n) for n in message.body[6:10])

            data = {
                'type': self.current_device_type, 
                'mdns_service': self.connected_mdns_service,
                'base_id': self.current_base_id, 
                'gateway_id': self.current_base_id,
                'tcm_version': tcm_sw_v, 
                'api_version': api_v
            }

            if hasattr(self._serial_bus, 'host') and hasattr(self._serial_bus, 'port'):
                data['address'] = f"{self._serial_bus._host}:{self._serial_bus._port}"

            self.app_bus.fire_event(AppBusEventType.ASYNC_TRANSCEIVER_DETECTED, data)
        
    def process_discovery_message(self, message: EltakoDiscoveryReply):
        
        if isinstance(message, EltakoDiscoveryReply):
            self.current_discovery_reply = message
            if self.current_base_id not in self.received_bus_device_discovery:
                self.received_bus_device_discovery[self.current_base_id] = []
            self.received_bus_device_discovery[self.current_base_id].append( message )
            
            dev = None
            for o in sorted_known_objects:
                if message.model[0:2] in o.discovery_names and (o.size is None or o.size == message.reported_size):
                    dev = o(message)
                    break
            if dev is None:
                dev = BusObject(message)
            
            async def get_all_sensors(): return []
            dev.get_all_sensors = get_all_sensors

            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Found device: {dev}", 'color':'grey'})
            self.app_bus.fire_event(AppBusEventType.DEVICE_SCAN_STATUS, 'DEVICE_DETECTED')
            self.app_bus.fire_event(AppBusEventType.ASYNC_DEVICE_DETECTED, {'device': dev, 'base_id': self.current_base_id, 'force_overwrite': False})


    def process_memory_response(self, message: EltakoMemoryResponse):
        if self.current_discovery_reply and isinstance(message, EltakoMemoryResponse):
            if self.current_discovery_reply not in self.received_bus_device_memory:
                self.received_bus_device_memory[self.current_discovery_reply] = []
            self.received_bus_device_memory[self.current_discovery_reply].append( message )
            
            if message.row == self.current_discovery_reply.memory_size-1:
                
                dev = None
                for o in sorted_known_objects:
                    if self.current_discovery_reply.model[0:2] in o.discovery_names and (o.size is None or o.size == self.current_discovery_reply.reported_size):
                        dev = o(self.current_discovery_reply)
                        break
                if dev is None:
                    dev = BusObject(self.current_discovery_reply)
                dev.memory = [r.value for r in self.received_bus_device_memory[self.current_discovery_reply]]
                
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Found device: {dev}", 'color':'grey'})
                self.app_bus.fire_event(AppBusEventType.DEVICE_SCAN_STATUS, 'DEVICE_DETECTED')
                asyncio.run( self.app_bus.fire_event(AppBusEventType.ASYNC_DEVICE_DETECTED, {'device': dev, 'base_id': self.current_base_id, 'force_overwrite': True}) )


    def connection_status_handler(self, connected: bool):
        self.app_bus.fire_event(AppBusEventType.CONNECTION_STATUS_CHANGE, {'connected': connected})


    def establish_serial_connection(self, serial_port:str, device_type:str) -> None:
        baudrate:int=57600
        delay_message:float=.1
        self.current_device_type = get_gateway_type_by_name(device_type)
        self.connected_mdns_service = None
        if device_type == GDN[GDT.EltakoFAMUSB]:
            baudrate = 9600
        elif device_type == GDN[GDT.EltakoFAM14]:
            delay_message = 0.001

        try:
            if not self.is_serial_connection_active():
                if device_type == GDN[GDT.LAN]:
                    ip_address = serial_port[:serial_port.rfind(':')]
                    port = int(serial_port[serial_port.rfind(':')+1:])
                    self.connected_mdns_service = self.gw_registry.find_mdns_service_by_ip_address(serial_port)
                    self._serial_bus = TCP2SerialCommunicator(ip_address, port,
                                                              callback=self._received_serial_event,
                                                              esp2_translation_enabled=True,
                                                              auto_reconnect=False
                                                              )

                elif device_type == GDN[GDT.ESP3]:
                    self._serial_bus = ESP3SerialCommunicator(serial_port, 
                                                              callback=self._received_serial_event,
                                                              esp2_translation_enabled=True,
                                                              auto_reconnect=False
                                                              )
                elif device_type == GDN[GDT.LAN_ESP2]:
                    ip_address = serial_port[:serial_port.rfind(':')]
                    port = int(serial_port[serial_port.rfind(':')+1:])
                    self.connected_mdns_service = self.gw_registry.find_mdns_service_by_ip_address(serial_port)
                    self._serial_bus = ESP2TCP2SerialCommunicator(ip_address, port,
                                                                  callback=self._received_serial_event,
                                                                  auto_reconnect=False)
                    
                else:
                    self._serial_bus = RS485SerialInterfaceV2(serial_port, 
                                                              baud_rate=baudrate, 
                                                              callback=self._received_serial_event, 
                                                              delay_message=delay_message,
                                                              auto_reconnect=False)
                self._serial_bus.start()
                self._serial_bus.is_serial_connected.wait(timeout=2)
                self._serial_bus.set_status_changed_handler(self.connection_status_handler)
                
                if not self._serial_bus.is_active():
                    self._serial_bus.stop()
                
                if self._serial_bus.is_active():
                    self.connected_gateway_type = device_type
                    if device_type in [GDN[GDT.LAN], GDN[GDT.LAN_ESP2] ]:
                        msg = f"TCP to Serial connection established. Server: {serial_port}"
                    else:
                        msg = f"Serial connection established. serial port: {serial_port}, baudrate: {baudrate}"
                    self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'color':'green'})
                    
                    if device_type == GDN[GDT.EltakoFAM14]:

                        def run():
                            asyncio.run( self._get_fam14_device_on_bus() )
                            self.app_bus.fire_event(
                                AppBusEventType.CONNECTION_STATUS_CHANGE, 
                                {'serial_port':  serial_port, 'baudrate': baudrate, 'connected': self._serial_bus.is_active()})
                            
                        t = threading.Thread(target=run)
                        t.start()
                    else:
                        if device_type in [ GDN[GDT.EltakoFAMUSB], GDN[GDT.ESP3], GDN[GDT.LAN] ]:
                            asyncio.run( self.async_create_gateway_device() )

                        self.app_bus.fire_event(
                                AppBusEventType.CONNECTION_STATUS_CHANGE, 
                                {'serial_port':  serial_port, 'baudrate': baudrate, 'connected': self._serial_bus.is_active()})
            
                else:
                    self.app_bus.fire_event(AppBusEventType.CONNECTION_STATUS_CHANGE, {'serial_port':  serial_port, 'baudrate': baudrate, 'connected': False})
                    if device_type == GDN[GDT.LAN]:
                        msg = f"Couldn't establish connection to {serial_port}! Try to restart device."
                    else:
                        msg = f"Establish connection for {device_type} on port {serial_port} failed! Device not ready."
                    self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})

        except Exception as e:
            self._serial_bus.stop()
            self.connected_gateway_type = None
            self.app_bus.fire_event(AppBusEventType.CONNECTION_STATUS_CHANGE, {'serial_port':  serial_port, 'baudrate': baudrate, 'connected': False})
            if device_type == GDN[GDT.LAN]:
                msg = f"Establish connection for {device_type} to {serial_port} failed! Retry later."
            else:
                msg = f"Establish connection for {device_type} on port {serial_port} failed!"
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
            logging.exception(msg, exc_info=True)
   

    async def async_create_gateway_device(self):

        await asyncio.sleep(2)

        await self._serial_bus.send_base_id_request()

        await self._serial_bus.send_version_request()                 

    

    def send_message(self, msg: EltakoMessage) -> None:
        asyncio.run( self._serial_bus.send(msg) ) 


    def stop_serial_connection(self) -> None:
        if self.is_serial_connection_active():
            self._serial_bus.stop()
            self._serial_bus._stop_flag.wait()

            time.sleep(0.5)

            self.app_bus.fire_event(AppBusEventType.CONNECTION_STATUS_CHANGE, {'connected': self._serial_bus.is_active()})
            if not self._serial_bus.is_active():
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Serial connection stopped.", 'color':'green'})
            self.current_base_id = None
            self.gateway_id = None
            self.connected_gateway_type = None
            self._serial_bus = None


    def kill_serial_connection_before_exit(self) -> None:
        if self.is_serial_connection_active():
            self._serial_bus.stop()


    def scan_for_devices(self, force_overwrite:bool=False) -> None:
        # if connected to FAM14
        if self.is_fam14_connection_active():
            
            t = threading.Thread(target=lambda: asyncio.run( self._scan_for_devices_on_bus(force_overwrite) )  )
            t.start()


    async def enumerate_bus(self) -> Iterator[BusObject]: # type: ignore
        """Search the bus for devices, yield bus objects for every match"""

        skip_until = 0

        for i in range(1, 256):
            try:
                if i > skip_until:
                    self.app_bus.fire_event(AppBusEventType.DEVICE_ITERATION_PROGRESS, i/256.0*100.0)
                    bus_object = await create_busobject(bus=self._serial_bus, id=i)
                    skip_until = i + bus_object.size -1
                    yield bus_object
            except TimeoutError:
                continue
            except Exception as e:
                msg = 'Cannot detect device'
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
                logging.exception(msg, exc_info=True)


    async def _get_fam14_device_on_bus(self, force_overwrite:bool=False) -> None:
        is_locked = False
        try:
            logging.debug(colored("Start scanning for devices", 'red'))

            self._serial_bus.set_callback( None )

            is_locked = (await locking.lock_bus(self._serial_bus)) == locking.LOCKED
            
            # first get fam14 and make it know to data manager
            fam14:FAM14 = await create_busobject(bus=self._serial_bus, id=255)
            self.current_base_id = await fam14.get_base_id()
            self.gateway_id = data_helper.a2s( (await fam14.get_base_id_in_int()) + 0xFF )
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Found device: {fam14}", 'color':'grey'})
            await self.app_bus.async_fire_event(AppBusEventType.ASYNC_DEVICE_DETECTED, {'device': fam14, 'base_id': self.current_base_id, 'force_overwrite': force_overwrite})

        except Exception as e:
            msg = 'Failed to load FAM14!!!'
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
            logging.exception(msg, exc_info=True)
            raise e
        finally:
            if is_locked:
                resp = await locking.unlock_bus(self._serial_bus)
            self._serial_bus.set_callback( self._received_serial_event )


    async def _scan_for_devices_on_bus(self, force_overwrite:bool=False) -> None:
        is_locked = False
        try:
            self.app_bus.fire_event(AppBusEventType.DEVICE_SCAN_STATUS, 'STARTED')
            
            msg = "Start scanning for devices"
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'color':'red'})

            self._serial_bus.set_callback( None )

            is_locked = (await locking.lock_bus(self._serial_bus)) == locking.LOCKED
            
            # first get fam14 and make it know to data manager
            fam14:FAM14 = await create_busobject(bus=self._serial_bus, id=255)
            logging.debug(colored(f"Found device: {fam14}",'grey'))
            await self.app_bus.async_fire_event(AppBusEventType.ASYNC_DEVICE_DETECTED, {'device': fam14, 'base_id': self.current_base_id, 'force_overwrite': force_overwrite})

            # iterate through all devices
            async for dev in self.enumerate_bus():
                try:
                    await dev.read_mem()
                    self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Found device: {dev}", 'color':'grey'})
                    self.app_bus.fire_event(AppBusEventType.DEVICE_SCAN_STATUS, 'DEVICE_DETECTED')
                    await self.app_bus.async_fire_event(AppBusEventType.ASYNC_DEVICE_DETECTED, {'device': dev, 'base_id': self.current_base_id, 'force_overwrite': force_overwrite})

                except TimeoutError:
                    logging.error("Read error, skipping: Device %s announces %d memory but produces timeouts at reading" % (dev, dev.discovery_response.memory_size))

            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': "Device scan finished.", 'color':'red'})
        except Exception as e:
            msg = 'Device scan failed!'
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
            logging.exception(msg, exc_info=True)
            raise e
        finally:
            # print("Unlocking the bus again")
            if is_locked:
                await locking.unlock_bus(self._serial_bus)

            self.app_bus.fire_event(AppBusEventType.DEVICE_SCAN_STATUS, 'FINISHED')
            self._serial_bus.set_callback( self._received_serial_event )

    def write_sender_id_to_devices(self, sender_id_list:dict={}):
        t = threading.Thread(target=lambda: asyncio.run( self.async_write_sender_id_to_devices(sender_id_list) )  )
        t.start()


    async def async_ensure_programmed(self, fam14_base_id_int:int, dev:BusObject, sender_id_list:dict):
        #HEATING = [FAE14SSR]
        if isinstance(dev, HasProgrammableRPS) or isinstance(dev, DimmerStyle):# or type(dev) in HEATING:
            for i in range(0,dev.size):
                device_ext_id_str = b2s( (fam14_base_id_int + dev.address+i).to_bytes(4,'big'))

                if device_ext_id_str in sender_id_list:
                    update_result = None
                    if 'sender' in sender_id_list[device_ext_id_str]:
                        sender_id_str = sender_id_list[device_ext_id_str]['sender']['id']
                        sender_eep_str = sender_id_list[device_ext_id_str]['sender']['eep']
                        sender_address = AddressExpression.parse(sender_id_str)
                        eep_profile = EEP.find(sender_eep_str)

                        # if type(dev) in HEATING:
                        #     # need to be at a special position 12 and 13
                        #     continue
                        #     mem_line:bytes = sender_address[0] + bytes((0, 65, 1, 0))
                        #     #TODO: NOT PROPERLY WORKING
                        #     await dev.write_mem_line(12 + i, mem_line)
                        # else:
                        retry = 3
                        exception = None
                        error_msg = {'msg': f'Failed to write sender id ({sender_id_str}) to device ({device_ext_id_str}). Will be retried up to {retry} times!',
                                        'log-level': 'ERROR', 
                                        'color': 'red'}

                        while retry > 0:
                            try:
                                update_result = await dev.ensure_programmed(i, sender_address, eep_profile)
                                retry=0
                                exception = None
                                time.sleep(0.2) # delay to avoid buffer overflow

                            except WriteError as e:
                                logging.exception(str(e))
                                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, error_msg)
                                retry -= 1
                                exception = e
                            except TimeoutError as e:
                                logging.exception(str(e))
                                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, error_msg)
                                retry -= 1
                                exception = e
                            except Exception as e:
                                logging.exception(str(e))
                                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, error_msg)
                                retry -= 1
                                exception = e
                                

                        if exception is not None:
                            raise exception
                                
                        if update_result is None:
                            msg = f'Update for device {type(dev).__name__} ({device_ext_id_str}) NOT supported.'
                        elif update_result == True:
                            msg = f'Updated Home Assistant sender id {sender_id_str} for eep {sender_eep_str} in device {type(dev).__name__} {device_ext_id_str}.'
                        else:
                            msg = f'Sender id {sender_id_str} for eep {sender_eep_str} in device {type(dev).__name__} {device_ext_id_str} already exists.'

                        self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'color':'grey'})
                        self.app_bus.fire_event(AppBusEventType.WRITE_SENDER_IDS_TO_DEVICES_STATUS, 'DEVICE_UPDATED')


    async def async_write_sender_id_to_devices(self, sender_id_list:dict={}): # 45056 = 0x00 00 B0 00
        if not self.is_fam14_connection_active():
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': "Cannot write HA sender ids to devices because you are not connected to FAM14.", 'color':'red'})
        else:
            try:
                self.app_bus.fire_event(AppBusEventType.WRITE_SENDER_IDS_TO_DEVICES_STATUS, 'STARTED')
                self._serial_bus.set_callback( None )

                # print("Sending a lock command onto the bus; its reply should tell us whether there's a FAM in the game.")
                time.sleep(0.2)
                await locking.lock_bus(self._serial_bus)
                
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': "Start writing Home Assistant sender ids to devices", 'color':'red'})

                # first get fam14 and make it know to data manager
                fam14:FAM14 = await create_busobject(bus=self._serial_bus, id=255)
                fam14_base_id_int = await fam14.get_base_id_in_int()
                fam14_base_id = b2s(await fam14.get_base_id_in_bytes())
                msg = f"Update devices on Bus (fam14 base id: {fam14_base_id})"
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'color':'grey'})

                # iterate through all devices
                async for dev in self.enumerate_bus():
                    await self.async_ensure_programmed(fam14_base_id_int, dev, sender_id_list)

                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': "Device scan finished.", 'color':'red'})
            except Exception as e:
                msg = 'Write sender id to devices failed!'
                self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': msg, 'log-level': 'ERROR', 'color': 'red'})
                logging.exception(msg, exc_info=True)
                raise e
            finally:
                # print("Unlocking the bus again")
                await locking.unlock_bus(self._serial_bus)

                self.app_bus.fire_event(AppBusEventType.WRITE_SENDER_IDS_TO_DEVICES_STATUS, 'FINISHED')
                self._serial_bus.set_callback( self._received_serial_event )


"""Constants for the Eltako integration."""
from enum import Enum
from strenum import StrEnum
import logging

from typing import Final

from homeassistant.const import *

DOMAIN: Final = "eltako"
DATA_ELTAKO: Final = "eltako"
DATA_ENTITIES: Final = "entities"
ELTAKO_GATEWAY: Final = "gateway"
CONF_GATEWAY_ADDRESS: Final ="address"
ELTAKO_CONFIG: Final = "config"
MANUFACTURER: Final = "Eltako"

ERROR_INVALID_GATEWAY_PATH: Final = "Invalid gateway path"
ERROR_NO_SERIAL_PATH_AVAILABLE: Final = "No serial path available. Try to reconnect your usb plug."
ERROR_NO_GATEWAY_CONFIGURATION_AVAILABLE: Final = "No gateway configuration available. Enter gateway into '/homeassistant/configuration.yaml'."

SIGNAL_RECEIVE_MESSAGE: Final = "receive_message"
SIGNAL_SEND_MESSAGE: Final = "send_message"
EVENT_BUTTON_PRESSED: Final = "btn_pressed"
EVENT_CONTACT_CLOSED: Final = "contact_closed"

LOGGER: Final = logging.getLogger(DOMAIN)

CONF_UNKNOWN: Final = "unknown"
CONF_REGISTERED_IN: Final = "registered_in"
CONF_COMMENT: Final = "comment"
CONF_EEP: Final = "eep"
CONF_DEVICE_ID: Final = "device_id"
CONF_SENSOR_KEY: Final = "sensor_key"
CONF_SENSOR_KEY_FUNCTION: Final = "sensor_key_func"
CONF_MEMORY_ENTRIES: Final = "memory_entries"
CONF_EXTERNAL_ID: Final = "external_id"
CONF_CONFIGURED_IN_DEVICES: Final = "configured_in_devices"
CONF_MEMORY_LINE: Final = "memory_line"
CONF_SWITCH_BUTTON: Final = "switch_button"
CONF_SENDER: Final = "sender"
CONF_SENSOR: Final = "sensor"
CONF_GERNERAL_SETTINGS: Final = "general_settings"
CONF_SHOW_DEV_ID_IN_DEV_NAME: Final = "show_dev_id_in_dev_name"
CONF_ENABLE_TEACH_IN_BUTTONS: Final = "enable_teach_in_buttons"
CONF_FAST_STATUS_CHANGE: Final = "fast_status_change"
GATEWAY_DEFAULT_NAME: Final = "EnOcean ESP2 Gateway"
CONF_GATEWAY: Final = "gateway"
CONF_GATEWAY_ID: Final = "gateway_id"
CONF_GATEWAY_DESCRIPTION: Final = "gateway_description"
CONF_BASE_ID: Final = "base_id"
CONF_DEVICE_TYPE: Final = "device_type"
CONF_SERIAL_PATH: Final = "serial_path"
CONF_CUSTOM_SERIAL_PATH: Final = "custom_serial_path"
CONF_MAX_TARGET_TEMPERATURE: Final = "max_target_temperature"
CONF_MIN_TARGET_TEMPERATURE: Final = "min_target_temperature"
CONF_ROOM_THERMOSTAT: Final = "thermostat"
CONF_COOLING_MODE: Final = "cooling_mode"

CONF_ID_REGEX: Final = "^([0-9a-fA-F]{2})-([0-9a-fA-F]{2})-([0-9a-fA-F]{2})-([0-9a-fA-F]{2})( (left|right))?$"
CONF_METER_TARIFFS: Final = "meter_tariffs"
CONF_TIME_CLOSES: Final = "time_closes"
CONF_TIME_OPENS: Final = "time_opens"
CONF_INVERT_SIGNAL: Final = "invert_signal"
CONF_VOC_TYPE_INDEXES: Final = "voc_type_indexes"


class LANGUAGE_ABBREVIATION(StrEnum):
    LANG_ENGLISH = 'en'
    LANG_GERMAN = 'de'


PLATFORMS: Final = [
    Platform.LIGHT,
    Platform.BINARY_SENSOR,
    Platform.SENSOR,
    Platform.SWITCH,
    Platform.COVER,
    Platform.CLIMATE,
    Platform.BUTTON,
]

class GatewayDeviceType(str, Enum):
    GatewayEltakoFAM14 = 'fam14'
    GatewayEltakoFGW14USB = 'fgw14usb'
    GatewayEltakoFAMUSB = 'fam-usb'     # ESP2 transceiver: https://www.eltako.com/en/product/professional-standard-en/three-phase-energy-meters-and-one-phase-energy-meters/fam-usb/
    EltakoFTD14 = 'ftd14'
    EnOceanUSB300 = 'enocean-usb300'
    EltakoFAM14 = 'fam14'
    EltakoFGW14USB = 'fgw14usb'
    EltakoFAMUSB = 'fam-usb'
    USB300 = 'enocean-usb300'
    ESP3 = 'esp3-gateway'
    LAN = 'lan'
    MGW_LAN = 'mgw-lan'
    EUL_LAN = 'EUL_LAN'
    LAN_ESP2 = "lan-gw-esp2"
    VirtualNetworkAdapter = 'esp2-netowrk-reverse-bridge'   # subtype of LAN_ESP2

    @classmethod
    def indexOf(cls, value):
        return list(cls).index(value)
    
    @classmethod
    def get_by_index(cls, index):
        return list(cls)[index]

    @classmethod
    def find(cls, value):
        for t in GatewayDeviceType:
            if t.value.lower() == value.lower():
                return t
        return None

    @classmethod
    def getValueByKeyOrValue(cls, input, default_value = None):
        if input in GatewayDeviceType.__members__:
            return GatewayDeviceType.__members__[input].value
        if input in [m.value for m in GatewayDeviceType]:
            return input
        return default_value

    @classmethod
    def is_transceiver(cls, dev_type) -> bool:
        return dev_type in [GatewayDeviceType.GatewayEltakoFAMUSB, GatewayDeviceType.EnOceanUSB300, GatewayDeviceType.USB300, GatewayDeviceType.ESP3, GatewayDeviceType.LAN, 
                            GatewayDeviceType.LAN_ESP2, GatewayDeviceType.MGW_LAN, GatewayDeviceType.EUL_LAN]

    @classmethod
    def is_bus_gateway(cls, dev_type) -> bool:
        return dev_type in [GatewayDeviceType.GatewayEltakoFAM14, GatewayDeviceType.GatewayEltakoFGW14USB,
                            GatewayDeviceType.EltakoFAM14, GatewayDeviceType.EltakoFAMUSB, GatewayDeviceType.EltakoFGW14USB]
    
    @classmethod
    def is_esp2_gateway(cls, dev_type) -> bool:
        return dev_type in [GatewayDeviceType.GatewayEltakoFAM14, GatewayDeviceType.GatewayEltakoFGW14USB, GatewayDeviceType.GatewayEltakoFAMUSB, 
                            GatewayDeviceType.EltakoFAM14, GatewayDeviceType.EltakoFAMUSB, GatewayDeviceType.EltakoFGW14USB, GatewayDeviceType.LAN_ESP2, 
                            GatewayDeviceType.VirtualNetworkAdapter]
    
    @classmethod
    def is_lan_gateway(cls, dev_type) -> bool:
        return dev_type in [GatewayDeviceType.LAN, GatewayDeviceType.LAN_ESP2, GatewayDeviceType.MGW_LAN, GatewayDeviceType.EUL_LAN, GatewayDeviceType.VirtualNetworkAdapter]


BAUD_RATE_DEVICE_TYPE_MAPPING: dict = {
    GatewayDeviceType.EltakoFAM14: 57600,
    GatewayDeviceType.EltakoFGW14USB: 57600,
    GatewayDeviceType.EltakoFAMUSB: 9600,
    GatewayDeviceType.USB300: 57600,
    GatewayDeviceType.ESP3: 57600,
}

GATEWAY_DISPLAY_NAMES = {
    GatewayDeviceType.EltakoFAM14: "FAM14 (ESP2)",
    GatewayDeviceType.EltakoFGW14USB: 'FGW14-USB (ESP2)',
    GatewayDeviceType.EltakoFAMUSB: 'FAM-USB (ESP2)',
    # GatewayDeviceType.USB300: 'ESP3 Gateway',
    GatewayDeviceType.ESP3: 'ESP3 Gateway',
    GatewayDeviceType.LAN: 'LAN Gateway (ESP3)',
    GatewayDeviceType.LAN_ESP2: "Home Assistant - Virtual Gateway Adapter",
}

def get_display_names():
    result = []
    for v in GATEWAY_DISPLAY_NAMES.values():
        if v not in result:
            result.append(v)
    return result

def get_gateway_type_by_name(device_type) -> GatewayDeviceType:
    for t, dn in GATEWAY_DISPLAY_NAMES.items():
        if device_type.lower() in dn.lower():
            return t
    return None


import os
from PIL import Image, ImageTk

# icon list
# https://commons.wikimedia.org/wiki/Comparison_of_icon_sets

class ImageGallery():

    # is needed to keep a reference to the images otherwise they will not displayed in the icons
    _images:dict[str:ImageTk.PhotoImage] = {}

    @classmethod
    def get_image(cls, filename, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        id = f"{filename}_{size[0]}_{size[0]}"
        if id not in cls._images:
            with Image.open(os.path.join(os.path.dirname(__file__), filename)) as img:
                if size is not None:
                    _img = img.resize(size, Image.LANCZOS)
                img = ImageTk.PhotoImage(_img)
                cls._images[id] = img

        return cls._images[id]

    @classmethod
    def get_eo_man_logo(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Faenza-system-search.png", size)

    @classmethod
    def get_blank(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("blank.png", size)

    @classmethod
    def get_ha_logo(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Home_Assistant_Logo.png", size)
    
    @classmethod
    def get_open_folder_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Oxygen480-status-folder-open.png", size)
    
    @classmethod
    def get_save_file_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Oxygen480-actions-document-save.png", size)
    
    @classmethod
    def get_save_file_as_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Oxygen480-actions-document-save-as.png", size)
    
    @classmethod
    def get_help_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Oxygen480-actions-help.png", size)
    
    @classmethod
    def get_paypal_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("paypal_icon.png", size)
    
    @classmethod
    def get_paypal_me_icon(cls, size:tuple[int:int]=(214,20)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("paypal_me_badge.png", size)
    
    @classmethod
    def get_about_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Breathe-about.png", size)

    @classmethod
    def get_info_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return cls.get_about_icon(size)

    @classmethod
    def get_github_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("github_icon.png", size)
    
    @classmethod
    def get_software_update_available_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Software-update-available.png", size)
    
    @classmethod
    def get_fam14_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("fam14.png", size)
    
    @classmethod
    def get_usb300_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("usb300.png", size)
    
    @classmethod
    def get_fam_usb_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("fam-usb2.png", size)
    
    @classmethod
    def get_ftd14_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("ftd14.png", size)
    
    @classmethod
    def get_fgw14_usb_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("fgw14-usb.png", size)
    
    @classmethod
    def get_wireless_network_in_color_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("wireless-network-colored.png", size)
    
    @classmethod
    def get_wireless_network_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("wireless-network-bw.png", size)
    
    @classmethod
    def get_wireless_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("wireless.png", size)
    
    @classmethod
    def get_send_mail(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("mail-send-receive.png", size)
    
    @classmethod
    def get_forward_mail(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("mail-forward.png", size)
    
    @classmethod
    def get_pct14_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("pct14.png", size)
    
    @classmethod
    def get_eul_gateway_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("EUL_device.png", size)
    
    @classmethod
    def get_mgw_piotek_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("mgw_piotek.png", size)
    
    @classmethod
    def get_clear_icon(cls, size:tuple[int:int]=(32,32)) -> ImageTk.PhotoImage:
        return ImageGallery.get_image("Oxygen480-actions-edit-clear.png", size)